""" 目的
    1. SignalP6 の結果を回収
    2. GTFprepフォーマットで出力
"""

import pandas as pd
import argparse
import sys
import csv

### output.gff3
#       ## gff-version 3
#       ENST00000261405.10      SignalP-6.0     signal_peptide  1       22      0.99975723      .       .       .
#       ENST00000216336.3       SignalP-6.0     signal_peptide  1       18      0.99973375      .       .       .

# |
# |
# |
# V

# #transcript_id	from	to	nt/aa		feature_name		source		attribute_field
# ENST00000261405.10	1	22	amino_acid	signal_peptide		SignalP-6.0	likelihood "0.99975723";


def parse_result(file):
    df = []
    with open(file, "r") as f:
        for line in f:
            if line.startswith("#"):
                continue
            cols = line.split()
            df.append([
                cols[0],	# 転写産物名
                int(cols[3]),	# from
                int(cols[4]),	# to
                float(cols[5]),	# likelihood
            ])
    return pd.DataFrame(df, columns=["Transcript", "from", "to", "likelihood"])


def filter_overlaped(df):
    df_filtered = df.loc[df.groupby('Transcript')['likelihood'].idxmax()]
    return df_filtered


def save_gtfprep(filtered_df, output_path, feature, source):
    """
    Parameters:
        filtered_df (pd.DataFrame): フィルタ済みデータフレーム
        output_path (str): 出力ファイルパス（TSV）
        feature (str): feature 列に使う固定文字列
        source (str): source 列に使う固定文字列
    """
    # attributes列を作成
    def format_attributes(row):
        return (
            'likelihood ' + '"'+str(row["likelihood"]).replace('"', '')+'"' + '; '
        )

    attributes = filtered_df.apply(format_attributes, axis=1)

    output_df = pd.DataFrame({
        "#Transcript": filtered_df["Transcript"],
        "tx_from": filtered_df["from"],
        "tx_to": filtered_df["to"],
        "amino_acid": "amino_acid",
        "feature": feature,
        "source": source,
        "attributes": attributes
    })

    # 保存
    output_df.to_csv(output_path, sep="\t", index=False, quoting=csv.QUOTE_NONE)
    print(">>> Output file: ", output_path)

def main():
    parser = argparse.ArgumentParser(description="Make gtfprep format from signalp results")
    parser.add_argument("signalp_output_gff3", help="Path to output.gff3")
    parser.add_argument("feature", help="feature name column")
    parser.add_argument("source", help="source name column")

    args = parser.parse_args()

    # 必須引数
    gff3 = args.signalp_output_gff3
    feature = args.feature
    source = args.source

    # メイン処理
    output_path = gff3 + ".gtfprep"
    df = parse_result(gff3)
    filtered_df = filter_overlaped(df)
    save_gtfprep(filtered_df, output_path, feature, source)

if __name__ == "__main__":
    main()




import subprocess
import argparse
import os

try:
    # 成功すれば、proteinToGenomeという名前が定義される
    import proteinToGenome
except ModuleNotFoundError:
    # 失敗した場合、絶対パスでインポートし、それをproteinToGenomeという名前に割り当てる
    import isoespy.proteinToGenome as proteinToGenome

### GTF
#	chr6	transdecoder	CDS	47712501	47712608	0.000000	+	.	gene_id "-"; transcript_id "ADGRF4_novel_tx_1"; gene_symbol "ADGRF4"; SPLICE_CATEGORY "novel_exon_combination";


### GTF prep format (transcript-based annotation format)
#	#Transcript		from	to	coordinate	feature			source		attributes
#	ADGRF4_novel_tx_1	252	500	amino_acid	protein_domain		pfam		domain_id "PF00002.29"; domain_name "7tm_2"; evalue "3.2e-33";
#	ADGRF4_novel_tx_1	199	242	amino_acid	protein_domain		pfam		domain_id "PF01825.27"; domain_name "GPS"; evalue; "4e-10";
#	ENST00000283303.3	252	500	amino_acid	protein_domain		pfam		domain_id "PF00002.29"; domain_name "7tm_2"; evalue "5.5e-33";
#	ENST00000283303.3	347	390	amino_acid	protein_domain		pfam		domain_id "PF01825.27"; domain_name "GPS"; evalue "5.4e-10";


def parse_meta(meta_file):
    ### 1. GTF meta
    # 初期値
    transcript_id = "transcript_id"
    cds = "CDS"
    exon = "exon"

    gtf_field = 0
    with open(meta_file) as f:
        for line in f:
            if line.startswith("[gtf]"):
                gtf_field = 1
                continue
            elif line.startswith("["):
                gtf_field = 0
            if gtf_field == 1:
                if line.strip() == "":
                    continue
                feature0 = line.split("=")[0].strip()[1:]
                feature1 = line.split("=")[1].strip()
                if feature0 == "transcript_id":
                    transcript_id = feature1
                if feature0 == "cds":
                    cds = feature1
                if feature0 == "exon":
                    exon = feature1

    return transcript_id, exon, cds


def parse_annotfile(annot_file):
    annot_data = {}
    feat_nbr = 0
    with open(annot_file) as f:
        for line in f:
            if line.startswith("#"):   # ヘッダー行は"#"で開始する
                continue
            feat_nbr += 1
            cols = line.strip().split('\t')

            tx_name = cols[0]
            featfrom = int(cols[1])
            featto = int(cols[2])
            coord = cols[3]
            feature = cols[4]
            source = cols[5]
            attribute = cols[6]
            if attribute == ".":
                attribute = ""

            # 処理
            if not tx_name in annot_data:
                annot_data[tx_name] = []
            annot_data[tx_name].append({"coordinate": coord, "from": featfrom, "to": featto, "feature": feature, "source": source, "attribute": attribute, "feature_id": feat_nbr})
    return annot_data


def parse_gtf(gtf_file, metadata_file, TRANSCRIPT_ID, EXON, CDS):
    """GTFファイルをパースし、転写産物ごとのcds情報を取得"""
    # TRANSCRIPT_ID : GTFファイル内におけるattributes 列の転写産物名フィールドのキー名
    # CDS           : GTFファイル内におけるfeature 列のCDS featureの名前
    # EXON          : GTFファイル内におけるfeature 列のexon featureの名前

    gtf_data = {}
    with open(gtf_file) as f:
        for line in f:
            if line.startswith("#"):  # コメント行はスキップ
                continue
            cols = line.strip().split('\t')
            if len(cols) < 9:
                continue
            
            chrom, source, feature, start, end, score, strand, frame, attributes = cols
            start, end = int(start), int(end)
            
            # transcript_idを取得
            attr_dict = {k.strip(): v.strip('"') for k, v in 
                         (item.split() for item in attributes.split(';') if item)}
            transcript_id = attr_dict.get(TRANSCRIPT_ID)

            # 処理
            if transcript_id:
                if feature == CDS:
                    if transcript_id not in gtf_data:
                        gtf_data[transcript_id] = {"chrom": chrom, "strand": strand, "exon": [], "cds": [], "score": score, "strand": strand, "frame": frame, "attributes": attributes}
                    gtf_data[transcript_id]["cds"].append((start, end))
                if feature == EXON:
                    if transcript_id not in gtf_data:
                        gtf_data[transcript_id] = {"chrom": chrom, "strand": strand, "exon": [], "cds": [], "score": score, "strand": strand, "frame": frame, "attributes": attributes}
                    gtf_data[transcript_id]["exon"].append((start, end))

    # exon を座標順に並び替え
    for tx in gtf_data:
        gtf_data[tx]["exon"] = sorted(gtf_data[tx]["exon"], key=lambda x: x[0])
    # CDS を座標順に並び替え
    for tx in gtf_data:
        gtf_data[tx]["cds"] = sorted(gtf_data[tx]["cds"], key=lambda x: x[0])

    # 塩基配列ベースのexon総延長(転写産物長)を計算する
    totallen_exon = {}
    for tx in gtf_data:
        exon_len = 0
        for i in gtf_data[tx]["exon"]:
            exon_len += i[1] - i[0] + 1
            totallen_exon[tx] = exon_len
    # 塩基配列ベースのCDS総延長を計算する
    totallen_cds = {}
    for tx in gtf_data:
        cds_len = 0
        for i in gtf_data[tx]["cds"]:
            cds_len += i[1] - i[0] + 1
            totallen_cds[tx] = cds_len
    for tx in gtf_data:
        if not tx in totallen_cds:
            totallen_cds[tx] = 0

    return gtf_data, totallen_exon, totallen_cds


def parse_frame(frame_file):
    frame_d = {}
    if frame_file == None:
        return {}

    else:
        with open(frame_file) as f:
            for line in f:
                if line.startswith("#"):
                    continue
                cols = line.strip().split("\t")
                frame_d[cols[0]] = int(cols[1])
        return frame_d


def ConvertAnnotationDataToGTF(annot_data, total_exon_d, total_cds_d, gtf_data, outfile_name, frame_d):
    for tx in annot_data:
        if tx in gtf_data:
            coord_gen_cds = gtf_data[tx]["cds"]
            coord_gen_exon = gtf_data[tx]["exon"]
            strand_gen = gtf_data[tx]["strand"]
            chrom = gtf_data[tx]["chrom"]
            score = gtf_data[tx]["score"]
            total_exon = total_exon_d[tx]
            total_cds = total_cds_d[tx]

            # for each feature for the transcript
            for feature_i in annot_data[tx]:
                coord_i = feature_i["coordinate"]
                from_i =  feature_i["from"]
                to_i = feature_i["to"]
                source = feature_i["source"]
                feature = feature_i["feature"]
                attributes_annot = feature_i["attribute"]
                feature_id = str(feature_i["feature_id"])

                # backmap
                if coord_i == "nucleotide":
                    backcoord_i = proteinToGenome.backmap_nuc(coord_gen_exon, tx, from_i, to_i, total_exon, strand_gen)		# ゲノム上にマップした塩基配列範囲
                elif coord_i == "amino_acid":
                    backcoord_i = proteinToGenome.backmap_aa(coord_gen_cds, tx, from_i, to_i, total_cds, strand_gen, frame_d)	# ゲノム上にマップしたアミノ酸配列範囲

                # GTFにfeature_iを追加
                for j in backcoord_i:
                    start, end = str(j[0]), str(j[1])
                    frame = "."

                    attributes_gtf = gtf_data[tx]["attributes"]                    
                    # attributeフィールドを統合 (annotation --> GTF)
                    attr_dict_annot = {k.strip(): v.strip('"') for k, v in (item.split() for item in attributes_annot.split(';') if item)}
                    attr_dict_gtf = {k.strip(): v.strip('"') for k, v in (item.split() for item in attributes_gtf.split(';') if item)}
                    for a in attr_dict_annot:
                        if not a in attr_dict_gtf:
                            # GTFのattributeフィールドに元々存在しないannotationファイル内のattributeを加える
                            attr_dict_gtf[a] = attr_dict_annot[a]
                    attributes_l = []
                    for a in attr_dict_gtf:
                        attributes_l.append(a+' '+'"'+attr_dict_gtf[a]+'"'+';')
                    attributes = ' '.join(attributes_l)
                    # 最後にfeature_id を追記
                    attributes += ' feature_id '+'"'+feature_id+'"'+';'
                    
                    with open(outfile_name, mode="a") as f:
                        f.write("\t".join([chrom, source, feature, start, end, score, strand_gen, frame, attributes]))
                        f.write("\n")


def outfileName(annotation_file, gtf_file):
    annot_base = os.path.basename(annotation_file)
    gtf_base = os.path.basename(gtf_file)
    annot_dir = os.path.dirname(annotation_file)
    return annot_dir + "/" + gtf_base + "_" + annot_base + "_" + "isoespy_makegtf.gtf"


def outfileCopy(gtf_file, outfile_name):
    subprocess.run(["rm", "-f", outfile_name])
    subprocess.run(["cp", gtf_file, outfile_name])


def separate_annotData(annot_data):
    """ GTF prep データをCDSの行とそれ以外のアノテーションの部分に分ける.
        CDSは先にGTF化する.
    """
    annot_data_cds = {}
    annot_data_else = {}
    for tx, features in annot_data.items():
        cds_list = [f for f in features if f.get("feature") == "CDS"]
        else_list = [f for f in features if f.get("feature") != "CDS"]
        if cds_list:
            annot_data_cds[tx] = cds_list
        if else_list:
            annot_data_else[tx] = else_list
    return annot_data_cds, annot_data_else


def isoespy_makegtf(gtfprep_file, gtf_file, meta_file, outfile_name, frame_file):
    # Metadata
    TRANSCRIPT_ID, EXON, CDS = parse_meta(meta_file)

    # frame_data
    frame_d = parse_frame(frame_file)

    # gtfprep data の分割
    annot_data = parse_annotfile(gtfprep_file)
    annot_data_cds, annot_data_else = separate_annotData(annot_data)

    # CDS --> GTF
    gtf_data, total_exon_d, total_cds_d = parse_gtf(gtf_file, meta_file, TRANSCRIPT_ID, EXON, CDS)
    # Get output file name
    if outfile_name == None:
        outfile_name = outfileName(gtfprep_file, gtf_file)
    # Initiate output file
    outfileCopy(gtf_file, outfile_name)
    # Make GTF
    ConvertAnnotationDataToGTF(annot_data_cds, total_exon_d, total_cds_d, gtf_data, outfile_name, frame_d)

    # non-CDS feature(s) --> GTF
    gtf_data, total_exon_d, total_cds_d = parse_gtf(outfile_name, meta_file, TRANSCRIPT_ID, EXON, CDS)
    # Make GTF
    ConvertAnnotationDataToGTF(annot_data_else, total_exon_d, total_cds_d, gtf_data, outfile_name, frame_d)


def main(args=None):
    # parse
    parser = argparse.ArgumentParser(description='isoespy_makegtf()')
    parser.add_argument('-p', '--gtfprep_file', required=True, type=str, default=None, help='GTF prep formatted file as input')
    parser.add_argument('-g', '--gtf_file', required=True, type=str, default=None, help='GTF file as input')
    parser.add_argument('-m', '--meta_file', required=True, type=str, default=None, help='meta file as input')
    parser.add_argument('-o', '--output', required=True, type=str, default=None, help='output file')
    parser.add_argument('-f', '--frame_file', type=str, default=None, help='Transcript translation frame file. Required when coordinate in GTF prep file contains amino_acid whose frame of coordinate may not be 0.')

    args = parser.parse_args()

    isoespy_makegtf(args.gtfprep_file, args.gtf_file, args.meta_file, args.output, args.frame_file)


if __name__ == "__main__":
    main()




import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.transforms as mtransforms
import copy
import pandas as pd
import seaborn as sns
import numpy as np
import re
import sys
import warnings
import argparse
import matplotlib.ticker as ticker
from matplotlib.cm import ScalarMappable
from matplotlib.colorbar import ColorbarBase
from matplotlib.colors import Normalize
from matplotlib.colors import LinearSegmentedColormap
import matplotlib.colors as mcolors
from collections import defaultdict

try:
    from intronCompression import intronCompression
except ModuleNotFoundError:
    try:
        from isoespy.intronCompression import intronCompression
    except ModuleNotFoundError:
        def intronCompression(model, ci):
            return model

# helper for attributes
def get_attr_candidate(attr_dict, keys):
    """Try multiple keys to find a value in attr_dict."""
    for k in keys:
        if k in attr_dict:
            return attr_dict[k]
    return None


# metadata parser
def parse_metadata(meta_data, gene):
    """Parse metadata file and return required dicts."""
    with open(meta_data) as f:
        lines = f.readlines()

    current_section = None
    sample_meta = dict()
    config_meta = dict()
    gtf_meta = dict()
    query = {"gene": [None, None], "tx": [None, None]}
    feature_meta = dict()

    # Hierarchical colors meta
    colors_meta = {
        "global": {
            "palette": None,
            "default_line": "gray",
            "default_text": "black",
            "default_tx": "#999999",
            "default_cds": "#3b73b9",
        },
        "transcripts": {},
        "exon": {"color": None},
        "cds": {"color": None},
        "de": {
            "box_group0": "#4daf4a",
            "box_group1": "#e41a1c"
        },
    }

    ctrl_trgt_table = dict()

    for line in lines:
        line = line.strip()
        if not line or line.startswith("!"):
            continue

        if line.startswith("[") and line.endswith("]"):
            current_section = line[1:-1]
            continue

        content = line.lstrip("#").strip() if line.startswith("#") else line

        if current_section == "sample":
            if line.startswith("#"):
                ctrl_trgt = content.split(",")
                for pair in ctrl_trgt:
                    if "=" in pair:
                        CT, var = pair.split("=")
                        ctrl_trgt_table[var.strip()] = CT.strip().lower()
            else:
                if "\t" in line:
                    i = line.split("\t")
                    sample_meta[i[0]] = i[1]
                elif "," in line:
                    i = line.split(",")
                    sample_meta[i[0]] = i[1]

        elif current_section == "config":
            if "=" in content:
                key, value = content.split("=", 1)
                key = key.strip()
                value = value.strip()
                if key == "qval":
                    config_meta[key] = float(value)
                elif key == "order":
                    config_meta[key] = [i.strip() for i in value.split(",")]
                elif key == "colors":
                    if ":" in value:
                        val1, val2 = value.split(":", 1)
                        val1 = val1.strip()
                        val2 = [i.strip() for i in val2.split(",")]
                        for tx in val2:
                            if tx not in colors_meta["transcripts"]:
                                colors_meta["transcripts"][tx] = val1

        elif current_section == "gtf":
            if "=" in content:
                key, value = content.split("=", 1)
                gtf_meta[key.strip()] = value.strip()

        elif current_section == "query":
            if "=" in content:
                key, value = content.split("=", 1)
                key = key.strip()
                value = value.strip()
                if key == "gene":
                    query[key] = [value, gene]
                elif key == "tx":
                    val_set = set([i.strip() for i in value.split(",")])
                    query[key][1] = val_set

        elif current_section == "features":
            if ":" in content:
                feature_class, feature_id = content.split(":", 1)
                feature_meta[feature_class.strip()] = feature_id.strip() if feature_id.strip() else None
            else:
                feature_meta[content] = None

        elif current_section == "colors":
            if "=" in content:
                key, value = content.split("=", 1)
                if key.strip() in colors_meta["global"]:
                    colors_meta["global"][key.strip()] = value.strip()

        elif current_section == "colors.transcripts":
            if "=" in content:
                key, value = content.split("=", 1)
                colors_meta["transcripts"][key.strip()] = value.strip()

        elif current_section == "colors.exon":
            if "=" in content:
                key, value = content.split("=", 1)
                if key.strip() == "color": colors_meta["exon"]["color"] = value.strip()
        
        elif current_section == "colors.cds":
            if "=" in content:
                key, value = content.split("=", 1)
                if key.strip() == "color": colors_meta["cds"]["color"] = value.strip()

        elif current_section == "colors.de":
            if "=" in content:
                key, value = content.split("=", 1)
                key = key.strip()
                if key in colors_meta["de"]:
                    colors_meta["de"][key] = value.strip()

    query["tx"][0] = gtf_meta.get("transcript_id")
    if query["tx"][1] == set():
        query["tx"][1] = None

    return sample_meta, config_meta, ctrl_trgt_table, gtf_meta, colors_meta, query, feature_meta


def get_scale_unit(data_range):
    target_size = data_range * 0.15
    order_of_magnitude = 10 ** int(np.floor(np.log10(target_size)))
    multipliers = [1, 2, 5, 10]
    selected_size = order_of_magnitude
    for m in multipliers:
        if order_of_magnitude * m <= target_size:
            selected_size = order_of_magnitude * m
        else:
            break
    if selected_size >= 1000:
        label = f"{int(selected_size/1000)} kb"
    else:
        label = f"{int(selected_size)} bp"
    return selected_size, label


def get_isoform_model(gtf_file, gtf_meta, tx_colors, query, colors_meta=None):
    transcripts = {}
    transcripts_CDS = {}
    target_gene = query["gene"][1]
    target_tx = query["tx"][1]

    with open(gtf_file) as gtf:
        for line in gtf:
            if line.startswith('#'): continue
            fields = line.strip().split('\t')
            if len(fields) < 9: continue
            chrom, source, feature, start, end, score, strand, frame, attributes = fields
            chrom = chrom.replace("chr","")
            
            attr_dict = {match.group(1): match.group(2) for match in re.finditer(r'(\S+)\s+"([^"]+)"', attributes)}
            
            transcript_id = attr_dict.get(gtf_meta['transcript_id'])
            line_gene = attr_dict.get(query["gene"][0])
            line_tx = attr_dict.get(query["tx"][0])
            
            if line_gene != target_gene: continue
            if target_tx is not None and line_tx not in target_tx: continue

            if feature == gtf_meta['exon']:
                if transcript_id not in transcripts:
                    direction = 1 if strand == "+" else -1
                    transcripts[transcript_id] = [[], direction, chrom]
                transcripts[transcript_id][0].append((int(start), int(end)))

            if feature == gtf_meta.get('cds', 'CDS'):
                if transcript_id not in transcripts_CDS:
                    direction = 1 if strand == "+" else -1
                    transcripts_CDS[transcript_id] = [[], direction, chrom]
                transcripts_CDS[transcript_id][0].append((int(start), int(end)))

    for tx in transcripts:
        transcripts[tx][0] = sorted(transcripts[tx][0], key=lambda x:x[0])
    for tx in transcripts_CDS:
        transcripts_CDS[tx][0] = sorted(transcripts_CDS[tx][0], key=lambda x:x[0])

    default_tx = "#B3C8CF"
    palette_name = None
    if colors_meta:
        default_tx = colors_meta["global"].get("default_tx", default_tx)
        palette_name = colors_meta["global"].get("palette", None)

    missing = [tx for tx in transcripts if tx not in tx_colors]
    if missing:
        if palette_name:
            pal = sns.color_palette(palette_name, len(missing))
            for tx, c in zip(missing, pal):
                tx_colors[tx] = mcolors.to_hex(c)
        else:
            for tx in missing:
                tx_colors[tx] = default_tx
    
    return transcripts, transcripts_CDS, tx_colors


# feature model helpers
def get_feature_model(gtf_file, gtf_meta, query, feature_meta):
    """Read positional functional features (GTF-based) with robust parsing."""
    print(f"DEBUG: Parsing features from {gtf_file}")
    ff_d = dict()

    target_gene = query["gene"][1]
    target_tx = query["tx"][1]
    
    # Candidate keys for gene name if primary key fails
    gene_keys = [query["gene"][0], 'gene_name', 'gene_symbol', 'gene_id', 'Name']
    gene_keys = [k for k in gene_keys if k]

    feature_count = 0
    with open(gtf_file) as gtf:
        for line in gtf:
            if line.startswith("#"): continue
            fields = line.strip().split("\t")
            if len(fields) < 9: continue
            chrom, source, feature, start, end, score, strand, frame, attributes = fields

            if feature in feature_meta:
                # Robust parsing 1: Regex
                attr_dict = {
                    match.group(1): match.group(2)
                    for match in re.finditer(r'(\S+)\s+"([^"]+)"', attributes)
                }
                # Robust parsing 2: Split by ; and =
                if not attr_dict:
                    for raw in attributes.strip().split(";"):
                        parts = raw.strip().split("=") 
                        if len(parts) == 2:
                            attr_dict[parts[0].strip()] = parts[1].strip().strip('"')

                transcript_id = get_attr_candidate(attr_dict, [gtf_meta.get("transcript_id"), "transcript_id", "transcript", "ID"])
                
                # Try finding gene name
                line_gene = get_attr_candidate(attr_dict, gene_keys)
                line_tx = get_attr_candidate(attr_dict, [query["tx"][0], "transcript_id", "transcript", "ID"])

                # Filtering logic
                match_gene = (line_gene == target_gene)
                match_tx = (target_tx is None) or (line_tx in target_tx)

                # Fallback: if transcript matches target list, accept even if gene name is missing
                if not match_gene and line_tx and target_tx and (line_tx in target_tx):
                    match_gene = True

                if not match_gene:
                    continue
                if not match_tx:
                    continue

                if transcript_id not in ff_d:
                    ff_d[transcript_id] = {feat: {} for feat in feature_meta}

                if feature_meta[feature] is None:
                    if ff_d[transcript_id][feature] == {}:
                        ff_d[transcript_id][feature]["NONAME"] = []
                    ff_d[transcript_id][feature]["NONAME"].append((int(start), int(end)))
                else:
                    feat_id = attr_dict.get(feature_meta[feature])
                    if feat_id:
                        if feat_id not in ff_d[transcript_id][feature]:
                            ff_d[transcript_id][feature][feat_id] = []
                        ff_d[transcript_id][feature][feat_id].append((int(start), int(end)))
                
                feature_count += 1

    for tx in ff_d:
        for feat in feature_meta:
            for ind in ff_d[tx][feat]:
                ff_d[tx][feat][ind] = sorted(ff_d[tx][feat][ind], key=lambda x: x[0])

    print(f"DEBUG: Total features accepted for gene {target_gene}: {feature_count}")
    return ff_d

def coordinate_adjustment(feat_names, Y_LOW, Y_UPP):
    out_d = dict()
    if feat_names == []:
        return out_d
    width = (Y_UPP - Y_LOW) / len(feat_names)
    for i in range(len(feat_names)):
        out_d[feat_names[i]] = (Y_UPP - width * (i + 1), Y_UPP - width * i)
    return out_d

def darken_color(color, factor=0.4):
    return tuple(max(0, c * factor) for c in color)

def overlaped_feature_indivisuals(DICT):
    def overlaps(r1, r2):
        return r1[0] <= r2[1] and r2[0] <= r1[1]

    graph = defaultdict(set)
    keys = list(DICT.keys())

    for i in range(len(keys)):
        for j in range(i + 1, len(keys)):
            k1, k2 = keys[i], keys[j]
            for r1 in DICT[k1]:
                for r2 in DICT[k2]:
                    if overlaps(r1, r2):
                        graph[k1].add(k2)
                        graph[k2].add(k1)
                        break
                else:
                    continue
                break

    visited = set()
    groups = []

    def dfs(node, group):
        visited.add(node)
        group.append(node)
        for neighbor in graph[node]:
            if neighbor not in visited:
                dfs(neighbor, group)

    for k in keys:
        if k not in visited:
            group = []
            dfs(k, group)
            groups.append(group)

    return groups


def formatting_isoform_model(transcripts_data, transcripts, annot):
    if annot == "exon":
        for transcript_id, exons in transcripts.items():
            if not exons[0]: continue
            start = min([i[0] for i in exons[0]])
            end = max(i[1] for i in exons[0])
            transcripts_data.append({'id': transcript_id, 'exons': exons[0], 'strand': exons[1], 'seq_region_name': exons[2], 'start': start, 'end': end})
    elif annot == "cds":
        for i in range(len(transcripts_data)):
            isomodel = transcripts_data[i]
            tx_id = isomodel['id']
            if tx_id in transcripts:
                transcripts_data[i]['cds'] = transcripts[tx_id][0]
            else:
                transcripts_data[i]['cds'] = []
    return transcripts_data


def get_expression_data(expression_file, transcripts_data, ctrl_trgt_table, sample_meta):
    df = pd.read_csv(expression_file, sep="\t", header=0)
    targets = [i['id'] for i in transcripts_data]
    filtered_df = df[df.iloc[:, 0].isin(targets)]
    GROUP_0, GROUP_1 = dict(), dict()
    
    for index, row in filtered_df.iterrows():
        tx_id = row.iloc[0]
        group_0, group_1 = [], []
        for sample in filtered_df.columns[1:]:
            if sample not in sample_meta: continue
            expression = row[sample]
            group_symbol = sample_meta[sample]
            
            if group_symbol not in ctrl_trgt_table: continue
            
            if ctrl_trgt_table[group_symbol] == 'control':
                group_0.append(expression)
            elif ctrl_trgt_table[group_symbol] == 'target':
                group_1.append(expression)

        GROUP_0[tx_id] = group_0
        GROUP_1[tx_id] = group_1

    for i in range(len(transcripts_data)):
        transcripts_data[i]['group0_exp'] = GROUP_0.get(transcripts_data[i]['id'], [])
        transcripts_data[i]['group1_exp'] = GROUP_1.get(transcripts_data[i]['id'], [])

    return transcripts_data


def get_det_data(det_file, transcripts_data):
    det_data = dict()
    with open(det_file) as f:
        for line in f:
            line = line.strip()
            if not line.startswith("#"):
                if "\t" in line:
                    col = line.split("\t")
                else:
                    col = line.split(",")
                if len(col) >= 3:
                    tx_id = col[0]
                    logFC = float(col[1])
                    qval = float(col[2])
                    det_data[tx_id] = [logFC, qval]

    for i in range(len(transcripts_data)):
        if transcripts_data[i]['id'] in det_data:
            transcripts_data[i]['det'] = det_data[transcripts_data[i]['id']]
        else:
            transcripts_data[i]['det'] = [0, 1.0]

    return transcripts_data


def prepare_de_ax1(transcripts_data, ax1, gene_name, tx_colors, colors_meta, tss_mode, model_features, feature_meta, ci=None, hide_scale=False):
    # AX1: 各アイソフォームを視覚化
    y_positions = []
    if not transcripts_data: return ax1
    
    MIN = min(transcripts_data, key=lambda x: x['start'])['start']
    MAX = max(transcripts_data, key=lambda x: x['end'])['end']
    transcripts_data = transcripts_data[::-1]
    all_tss = []
    
    if model_features == {}:
        feat_names = []
        feat_colors = []
    else:
        all_found_keys = set()
        for tx in model_features:
            all_found_keys.update(model_features[tx].keys())
        
        feat_names = [k for k in feature_meta.keys() if k in all_found_keys]
        feat_colors = sns.color_palette("husl", len(feat_names))

    gcol = colors_meta.get("global", {})
    line_color = gcol.get("default_line", "gray")
    text_color = gcol.get("default_text", "black")
    default_tx_color = colors_meta.get("global", {}).get("default_tx", "#999999")
    default_cds_color = colors_meta.get("global", {}).get("default_cds", "#3b73b9")

    exon_override = colors_meta.get("exon", {}).get("color", None)
    cds_override = colors_meta.get("cds", {}).get("color", None)
    user_tx_colors = colors_meta.get("user_transcripts", {})

    for i, transcript_data in enumerate(transcripts_data):
        y_positions.append(i)
        start = transcript_data['start']
        end = transcript_data['end']
        tx_name = transcript_data['id']
        strand = transcript_data['strand']

        arrow_direction = "right" if strand == 1 else "left"
        y_pos = i
        
        ax1.annotate('', xy=(end, y_pos), xytext=(start, y_pos), arrowprops=dict(arrowstyle="-", color=line_color, lw=1))
        
        interval = max(1, (MAX-MIN)//50)
        x_positions = np.arange(start, end, interval)
        x_positions = x_positions[1:-1]
        
        marker_shape = ">" if arrow_direction == "right" else "<"
        for x in x_positions:
            ax1.scatter(x, y_pos, marker=marker_shape, color=line_color, s=10)

        if strand == 1:
            all_tss.append(start)
        else:
            all_tss.append(end)

        # Exon
        if tx_name in user_tx_colors:
            current_exon_color = user_tx_colors[tx_name]
        elif exon_override:
            current_exon_color = exon_override
        else:
            current_exon_color = tx_colors.get(tx_name, default_tx_color)

        # CDS
        base_color = tx_colors.get(tx_name, default_tx_color)
        if cds_override:
            current_cds_color = cds_override
        elif tx_name in user_tx_colors:
            current_cds_color = user_tx_colors[tx_name]
        elif base_color == default_tx_color:
            current_cds_color = default_cds_color
        else:
            current_cds_color = base_color

        for exon in transcript_data['exons']:
            exon_start = exon[0]
            exon_end = exon[1]
            ax1.add_patch(patches.Rectangle((exon_start, i - 0.1), exon_end - exon_start, 0.2, color=current_exon_color))

        for cds in transcript_data['cds']:
            cds_start = cds[0]
            cds_end = cds[1]
            ax1.add_patch(patches.Rectangle((cds_start, i - 0.2), cds_end - cds_start, 0.4, color=current_cds_color))

        if tx_name not in model_features:
            continue

        Y_LOW = i - 0.8
        Y_UPP = i - 0.22 
        coord_d = coordinate_adjustment(feat_names, Y_LOW, Y_UPP)

        for k_feat, feat in enumerate(feat_names):
            if feat not in model_features[tx_name]: continue
            y_low, y_upp = coord_d[feat]
            if not model_features[tx_name][feat]: continue
            
            plot_order = sorted(
                model_features[tx_name][feat],
                key=lambda k: sum(e - s for s, e in model_features[tx_name][feat][k]),
                reverse=True,
            )
            plot_groups = overlaped_feature_indivisuals(model_features[tx_name][feat])
            plot_groups_cnt = [-1 for _ in range(len(plot_groups))]

            for ind in plot_order:
                idx = 0
                for g in range(len(plot_groups)):
                    if ind in plot_groups[g]:
                        plot_groups_cnt[g] += 1
                        idx = g
                        break

                y_low2 = y_low - 0.009 * plot_groups_cnt[idx]
                y_upp2 = y_upp - 0.009 * plot_groups_cnt[idx]
                
                crush_rate = 0.5
                full_height = y_upp2 - y_low2
                box_height = full_height * crush_rate
                y_center = (y_upp2 + y_low2) / 2
                y_start = y_center - (box_height / 2)

                for (start_k, end_k) in model_features[tx_name][feat][ind]:
                    ax1.add_patch(
                        patches.Rectangle(
                            (start_k, y_start),
                            end_k - start_k,
                            box_height,
                            color=feat_colors[k_feat],
                        )
                    )

                leftmost_x = model_features[tx_name][feat][ind][0][0]
                rightmost_x = model_features[tx_name][feat][ind][-1][1]
                ax1.add_patch(
                    patches.Rectangle(
                        (leftmost_x, y_start),
                        rightmost_x - leftmost_x,
                        box_height,
                        linewidth=0.5,
                        edgecolor=darken_color(feat_colors[k_feat]),
                        facecolor="none",
                    )
                )

    if tss_mode:
        unique_tss = sorted(set(all_tss))
        ymin = -0.3
        ymax = len(transcripts_data) - 0.8
        for x in unique_tss:
            ax1.vlines(x, ymin=ymin, ymax=ymax, colors=line_color, linestyles=':', linewidth=0.8, alpha=0.4)

    # === スケールバー ===
    if not hide_scale:
        data_width = MAX - MIN
        scale_size, scale_label = get_scale_unit(data_width)
        bar_y = - 0.8
        text_y = bar_y - 0.05
        bar_end = MAX
        bar_start = MAX - scale_size
        ax1.plot([bar_start, bar_end], [bar_y, bar_y], color="#3B5894", linewidth=1, clip_on=False)
        cap_height = 0.1
        ax1.plot([bar_start, bar_start], [bar_y - cap_height/2, bar_y + cap_height/2], color="#3B5894", linewidth=1, clip_on=False)
        ax1.plot([bar_end, bar_end], [bar_y - cap_height/2, bar_y + cap_height/2], color="#3B5894", linewidth=1, clip_on=False)
        ax1.text((bar_start + bar_end) / 2, text_y, scale_label, ha='center', va='top', fontsize=8, color="#3B5894")
        if ci is not None:
            note_text = "Note: Introns compressed"
            ax1.text(bar_end, bar_y + 0.1, note_text, ha='right', va='bottom', fontsize=8, color="#3B5894", style='italic')
    # ========================

    space = int((MAX-MIN)/20)
    ax1.set_xlim(MIN - 2*space, MAX + space)
    ax1.set_ylim(-1.0, len(transcripts_data) - 0.5)
    
    chrom_label = transcripts_data[0]["seq_region_name"]
    chrom_norm = chrom_label.lower()
    if chrom_norm.startswith("chr"):
        chrom_norm = chrom_label[3:]
    else:
        chrom_norm = chrom_label

    ax1.set_xlabel(f'Chr{chrom_norm}', color=text_color)
    ax1.set_title(gene_name, color=text_color)
    ax1.set_yticks(y_positions)
    transcripts = [i['id'] for i in transcripts_data]
    ax1.set_yticklabels(transcripts, color=text_color)

    N = len(transcripts_data) - 1
    if feat_names:
        coord = coordinate_adjustment(feat_names, Y_LOW=N - 0.8, Y_UPP=N - 0.22)
        for k_feat, feat in enumerate(feat_names):
            y_low, y_upp = coord[feat]
            ax1.text(
                MIN - 1.8 * space,
                (y_upp + y_low) / 2,
                feat,
                fontsize=9,
                color=darken_color(feat_colors[k_feat], factor=0.6),
                ha="left",
                va="center",
            )

    for label in ax1.get_xticklabels():
        label.set_rotation(15)
        label.set_ha('right')
        label.set_color(text_color)
        
    return ax1


def prepare_ax1_xaxis(ax1, ci, x_min, x_max, x_min_eff, x_max_eff):
    if ci == None:
        ax1.xaxis.set_major_formatter(ticker.FormatStrFormatter('%d'))
    else:
        ax1.set_xticks([x_min_eff, x_max_eff])
        ax1.set_xticklabels([str(x_min), str(x_max)])
    return ax1


def prepare_de_ax2(transcripts_data, ax2, config_meta, colors_meta, outliers, group0_label, group1_label):
    transcripts_data = transcripts_data[::-1]
    transcripts = [i['id'] for i in transcripts_data]
    labels = []
    data = []
    
    for i in range(len(transcripts_data)):
        tx_id = transcripts_data[i]['id']
        data.append(transcripts_data[i]['group1_exp']) 
        data.append(transcripts_data[i]['group0_exp'])
        labels.append(f'{tx_id}_1')
        labels.append(f'{tx_id}_0')

    positions = []
    for i, name in enumerate(transcripts):
         positions.extend([i-0.12, i+0.12])

    boxplot = ax2.boxplot(data, positions=positions, vert=False, patch_artist=True, widths=0.2, showfliers=outliers)

    c_group0 = colors_meta['de'].get('box_group0', '#4daf4a')
    c_group1 = colors_meta['de'].get('box_group1', '#e41a1c')
    
    box_colors = [c_group1 if idx % 2 == 0 else c_group0 for idx in range(len(data))]
    
    for patch, color in zip(boxplot['boxes'], box_colors):
        patch.set_facecolor(color)

    for median in boxplot['medians']:
        median.set_color('black')

    ax2.set_ylim(-1.0, len(transcripts_data) - 0.5) 
    ax2_xlabel = "Expression level"
    ax2.set_yticklabels([])
    ax2.tick_params(axis='y', length=0)
    ax2.set_xlabel(ax2_xlabel)

    legend_handles = [
        patches.Patch(facecolor=c_group1, edgecolor='black', label=group1_label),
        patches.Patch(facecolor=c_group0, edgecolor='black', label=group0_label),
    ]
    ax2.legend(handles=legend_handles, frameon=False)
    return ax2


def prepare_de_ax3(transcripts_data, qval_threshold, ax3):
    # AX3: ヒートマップ (q値) + q値/logFCのテキスト表示
    
    # === ax1と順序を合わせる ===
    transcripts_data = transcripts_data[::-1]
    
    # === 全体の描画範囲を等分する ===
    N = len(transcripts_data)
    Y_MIN = -1.0
    Y_MAX = N - 0.5
    TOTAL_HEIGHT = Y_MAX - Y_MIN
    UNIT_HEIGHT = TOTAL_HEIGHT / N 

    def color_mapping(nlogqval):
        if nlogqval > 20: return 1
        else: return 0.3 + nlogqval*0.035
    
    DET = list()
    for i in range(len(transcripts_data)):
        DET.append([transcripts_data[i]['id']] + transcripts_data[i]['det'])

    warm_cmap = plt.cm.Reds
    cold_cmap = plt.cm.Blues

    for i, val in enumerate(DET):
        logFC = float(DET[i][1]) # floatに変換
        qval = float(DET[i][2])  # floatに変換
        
        # 有意かどうか判定
        is_significant = qval < float(qval_threshold)

        # 色の決定
        if is_significant:
            if logFC > 0:
                color = warm_cmap(color_mapping(-np.log10(qval)))
            else:
                color = cold_cmap(color_mapping(-np.log10(qval)))
        else:
                color = (0.5, 0.5, 0.5, 0.5)

        # ヒートマップのバーを描画
        y_bottom = Y_MIN + (i * UNIT_HEIGHT)
        rect = patches.Rectangle(
            (0, y_bottom),
            0.1,  # バーの幅
            UNIT_HEIGHT,
            facecolor=color, 
            edgecolor='black', 
            linewidth=2
        )
        ax3.add_patch(rect)

        # === 【テキスト描画ロジック】 ===
        
        # 1. q値の文字列作成
        if qval < 0.001:
            # 非常に小さい値は指数表記 (例: q=1.2e-05)
            text_str = f"q={qval:.1e}\n"
        else:
            # 通常は小数第3位まで (例: q=0.052)
            text_str = f"q={qval:.3f}\n"
            
        # 2. 有意な場合のみ logFC を追加
        logfc_str = f"{logFC:.2f}".replace('-', '–')
        if is_significant:
            text_str += f"logFC={logfc_str}"
        else:
            text_str += f"logFC=Not Sig."

        # 3. テキストを描画
        ax3.text(
            0.15,  # バーの少し右
            y_bottom + (UNIT_HEIGHT / 2), 
            text_str,
            ha='left',      # 左揃え
            va='center',    # 上下中央揃え
            fontsize=10,     # 文字が長くなるので少し小さめに設定
            color='black'
        )
        # ==============================

    # カラーバーの描画設定（変更なし）
    reds_colors = [plt.cm.Reds(x) for x in np.linspace(0.3, 1, 128)]
    blues_colors = [plt.cm.Blues(x) for x in np.linspace(0.3, 1, 128)]
    combined_colors = blues_colors[::-1] + reds_colors
    combined_cmap = LinearSegmentedColormap.from_list("CombinedMap", combined_colors)
    norm = Normalize(vmin=0, vmax=1)
    sm = ScalarMappable(cmap=combined_cmap, norm=norm)
    sm.set_array([])
    cbar = plt.colorbar(sm, ax=ax3, orientation='vertical', fraction=0.05, pad=0.01, shrink=0.8)
    cbar.set_label('-log10(q)', rotation=90)
    cbar.set_ticks(np.linspace(0, 1, 5))
    tick_labels = [">20", "10", "0", "10", ">20"]
    cbar.ax.set_yticklabels(tick_labels)

    ax3.set_xlim(0, 1)
    ax3.set_ylim(Y_MIN, Y_MAX)
    ax3.axis('off')
    return ax3


def reorder(transcripts_data, meta_data):
    if not 'order' in meta_data:
        return transcripts_data
    else:
        tmp = []
        order = meta_data['order']
        for tx in order:
            for i in transcripts_data:
                if i['id'] == tx:
                    tmp.append(i)
                    break
        return tmp


def plot_isoespy_de(
        transcripts_data,
        config_meta,
        gene_name,
        tx_colors,
        colors_meta,
        ci,
        x_min,
        x_max,
        x_min_eff,
        x_max_eff,
        outliers,
        group0_label,
        group1_label,
        tss_mode,
        model_features,
        feature_meta, # Added
        output_file=None,
        dpi=300,
        q_thresh=0.05,
        hide_scale=False):
    
    transcripts_data = reorder(transcripts_data, config_meta)
    
    if not transcripts_data:
        print("No transcripts found for plotting.")
        return

    fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(20, 5), gridspec_kw={'width_ratios': [10, 6, 2]})

    ax1 = prepare_de_ax1(transcripts_data, ax1, gene_name, tx_colors, colors_meta, tss_mode, model_features, feature_meta, ci, hide_scale)
    ax1 = prepare_ax1_xaxis(ax1, ci, x_min, x_max, x_min_eff, x_max_eff)

    ax2 = prepare_de_ax2(transcripts_data, ax2, config_meta, colors_meta, outliers, group0_label, group1_label)

    ax3 = prepare_de_ax3(transcripts_data, q_thresh, ax3)

    plt.subplots_adjust(wspace=0.02)

    if output_file:
        try:
            fig.savefig(output_file, dpi=dpi, bbox_inches='tight')
            print(f"Figure successfully saved to {output_file} with DPI={dpi}")
        except Exception as e:
            print(f"Error saving figure to file: {e}")
            plt.show()
    else:
        plt.show()


def moved_data_for_exons_cds(main_data, ci):
    model = {}
    for tx_data in main_data:
        tx_name = tx_data['id']
        model[tx_name+"_exons"] = tx_data['exons']
        model[tx_name+"_cds"] = tx_data['cds']

    model_compressed = intronCompression(model, ci)

    startend_d = {}
    for tx_data in main_data:
        tx_name = tx_data['id']
        if tx_name+"_exons" in model_compressed:
            exons = model_compressed[tx_name+"_exons"]
            if exons:
                start = exons[0][0]
                end = exons[-1][1]
                startend_d[tx_name] = {"start": start, "end": end}
                tx_data['exons'] = exons
                tx_data['cds'] = model_compressed[tx_name+"_cds"]
                tx_data['start'] = start
                tx_data['end'] = end
    return main_data

def moved_data_for_features(model_features, main_data, ci):
    model = {}
    for tx_data in main_data:
        tx_name = tx_data["id"]
        model[tx_name + "_exons"] = tx_data["exons"]
        model[tx_name + "_cds"] = tx_data["cds"]

    for tx_name in model_features:
        for feat_name in model_features[tx_name]:
            for id_name in model_features[tx_name][feat_name]:
                model[tx_name + "_" + feat_name + "_" + id_name] = model_features[tx_name][feat_name][id_name]

    model_compressed = intronCompression(model, ci)

    for tx_name in model_features:
        for feat_name in model_features[tx_name]:
            for id_name in model_features[tx_name][feat_name]:
                key = tx_name + "_" + feat_name + "_" + id_name
                if key in model_compressed:
                    model_features[tx_name][feat_name][id_name] = model_compressed[key]

    return model_features


def isoespy_de(gene, gtf_data, expression_data, det_data, meta_data, ci, outliers, group0_label, group1_label, tss_mode, output_file=None, dpi=300, q_thresh=0.05, hide_scale=False, show_features=False):
    sample_meta, config_meta, ctrl_trgt_table, gtf_meta, colors_meta, query, feature_meta = parse_metadata(meta_data, gene)

    tx_colors = colors_meta["transcripts"]
    model_exon, model_cds, tx_colors = get_isoform_model(gtf_data, gtf_meta, tx_colors, query, colors_meta)
    
    # Feature model (Conditional)
    if show_features:
        model_features = get_feature_model(gtf_data, gtf_meta, query, feature_meta)
    else:
        model_features = {}

    transcripts_main_data = list()
    transcripts_main_data = formatting_isoform_model(transcripts_main_data, model_exon, annot="exon")
    transcripts_main_data = formatting_isoform_model(transcripts_main_data, model_cds, annot="cds")
    
    if not transcripts_main_data:
        print(f"No transcripts found for gene: {gene}")
        return

    x_min = min(transcripts_main_data, key=lambda x: x['start'])['start']
    x_max = max(transcripts_main_data, key=lambda x: x['end'])['end']

    if ci != None:
        copied_data = copy.deepcopy(transcripts_main_data)
        transcripts_main_data = moved_data_for_exons_cds(transcripts_main_data, ci)
        if show_features:
            model_features = moved_data_for_features(model_features, copied_data, ci)
    
    x_min_eff = min(transcripts_main_data, key=lambda x: x['start'])['start']
    x_max_eff = max(transcripts_main_data, key=lambda x: x['end'])['end']

    transcripts_main_data = get_expression_data(expression_data, transcripts_main_data, ctrl_trgt_table, sample_meta)
    transcripts_main_data = get_det_data(det_data, transcripts_main_data)

    plot_isoespy_de(
        transcripts_main_data,
        config_meta,
        gene,
        tx_colors,
        colors_meta,
        ci,
        x_min,
        x_max,
        x_min_eff,
        x_max_eff,
        outliers,
        group0_label,
        group1_label,
        tss_mode,
        model_features,
        feature_meta,
        output_file,
        dpi,
        q_thresh,
        hide_scale
    )


def process_ci(ci):
    if ci is None: return None
    if isinstance(ci, str):
        try:
            ci = float(ci)
        except ValueError:
            warnings.warn(f"Warning: ci should be a float/int, but received string '{ci}'")
            return None
    return ci


def main(args=None):
    parser = argparse.ArgumentParser(description='isoespy_de()')
    parser.add_argument('-gene', '--gene_name', required=True, type=str, default=None, help='Gene name')
    parser.add_argument('-gtf', '--gtf_data', required=True, type=str, default=None, help='GTF file')
    parser.add_argument('-exp', '--expression_data', required=True, type=str, default=None, help='Expression data')
    parser.add_argument('-det', '--det_data', required=True, type=str, default=None, help='Differential expression data')
    parser.add_argument('-meta', '--meta_data', required=True, type=str, default=None, help='metadata')
    parser.add_argument('-ci', '--compress_introns', default=None, help='intron compression parameter')
    parser.add_argument('--show_outliers', action="store_true", help="Show outliers in the boxplot (default: hide)")
    parser.add_argument('--group0_label', default='0', help='Label of control (eg. Nontumor)')
    parser.add_argument('--group1_label', default='1', help='Label of target (eg. HCC)')
    parser.add_argument('-tss', '--tss_line', action="store_true", help="Show TSS support lines")
    parser.add_argument('-qval', '--q_thresh', type=float, default=0.05, help='q-value/FDR threshold for coloring significant isoforms')
    parser.add_argument("-o", "--output_file", type=str, default=None, help="Save figure to file (e.g., plot.pdf or plot.png)")
    parser.add_argument("--dpi", type=int, default=300, help="Resolution for raster image formats (e.g., PNG) in DPI")
    parser.add_argument("--hide_scale", action="store_true", help="Hide scale bar")
    parser.add_argument("-sf", "--show_features", action="store_true", help="Show genomic features (e.g. domains) in the plot")

    args = parser.parse_args()
    gene = args.gene_name
    gtf_data = args.gtf_data
    expression_data = args.expression_data
    det_data = args.det_data
    meta_data = args.meta_data
    ci = process_ci(args.compress_introns)
    outliers = args.show_outliers
    group0_label = args.group0_label
    group1_label = args.group1_label
    tss_mode = args.tss_line
    q_thresh = args.q_thresh
    output_file = args.output_file
    dpi = args.dpi
    hide_scale = args.hide_scale
    show_features = args.show_features

    isoespy_de(gene, gtf_data, expression_data, det_data, meta_data, ci, outliers, group0_label, group1_label, tss_mode, output_file, dpi, q_thresh, hide_scale, show_features)


if __name__ == '__main__':
    main()




import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.transforms as mtransforms
import copy
import pandas as pd
import seaborn as sns
import numpy as np
import re
import sys
import warnings
import argparse
import matplotlib.ticker as ticker
import matplotlib.colors as mcolors
from collections import defaultdict
from matplotlib.cm import ScalarMappable
from matplotlib.colorbar import ColorbarBase
from matplotlib.colors import Normalize
from matplotlib.colors import LinearSegmentedColormap
from mpl_toolkits.axes_grid1.inset_locator import inset_axes

try:
    from intronCompression import intronCompression
except ModuleNotFoundError:
    try:
        from isoespy.intronCompression import intronCompression
    except ModuleNotFoundError:
        def intronCompression(model, ci): return model


# metadata parser
def parse_metadata(meta_data, gene):
    """Parse metadata file and return required dicts."""
    with open(meta_data) as f:
        lines = f.readlines()

    current_section = None
    sample_meta = dict()
    config_meta = dict()
    gtf_meta = dict()
    query = {"gene": [None, None], "tx": [None, None]}
    feature_meta = dict()

    # Hierarchical colors meta
    colors_meta = {
        "global": {
            "palette": None,
            "default_line": "gray",
            "default_text": "black",
            "default_tx": "#999999",
            "default_cds": "#3b73b9",
        },
        "transcripts": {},
        "exon": {"color": None},
        "cds": {"color": None},
        "ff": {
            "continuous": {"cmap": "viridis"},
            "categorical": {"palette": "tab10", "colors": {}},
            "binary": {"0": "#cccccc", "1": "#000000"},
        },
    }

    ctrl_trgt_table = dict()

    for line in lines:
        line = line.strip()
        # コメント行の処理 (!で始まる行は無視)
        if not line or line.startswith("!"):
            continue

        if line.startswith("[") and line.endswith("]"):
            current_section = line[1:-1]
            continue

        # パラメータ行の解析 (#key=value または key=value)
        # コメント記号 # を除去して key, value を取得
        content = line.lstrip("#").strip() if line.startswith("#") else line

        # ---------------- sample ----------------
        if current_section == "sample":
            # sampleセクションは #始まりが target definition, それ以外が sample table
            if line.startswith("#"):
                ctrl_trgt = content.split(",")
                for pair in ctrl_trgt:
                    if "=" in pair:
                        CT, var = pair.split("=")
                        ctrl_trgt_table[var.strip()] = CT.strip().lower()
            else:
                if "\t" in line:
                    i = line.split("\t")
                    sample_meta[i[0]] = i[1]
                elif "," in line:
                    i = line.split(",")
                    sample_meta[i[0]] = i[1]

        # ---------------- config ----------------
        elif current_section == "config":
            if "=" in content:
                key, value = content.split("=", 1)
                key = key.strip()
                value = value.strip()
                if key == "qval":
                    config_meta[key] = float(value)
                elif key == "order":
                    config_meta[key] = [i.strip() for i in value.split(",")]
                elif key == "colors":
                    # legacy style: "#colors = #HEX: tx1,tx2" (support but hierarchical wins)
                    if ":" in value:
                        val1, val2 = value.split(":", 1)
                        val1 = val1.strip()
                        val2 = [i.strip() for i in val2.split(",")]
                        for tx in val2:
                            if tx not in colors_meta["transcripts"]:
                                colors_meta["transcripts"][tx] = val1

        # ---------------- gtf ----------------
        elif current_section == "gtf":
            if "=" in content:
                key, value = content.split("=", 1)
                gtf_meta[key.strip()] = value.strip()

        # ---------------- query ----------------
        elif current_section == "query":
            if "=" in content:
                key, value = content.split("=", 1)
                key = key.strip()
                value = value.strip()
                if key == "gene":
                    query[key] = [value, gene]
                elif key == "tx":
                    val_set = set([i.strip() for i in value.split(",")])
                    query[key][1] = val_set

        # ---------------- features ----------------
        elif current_section == "features":
            # features usually defined like "#Protein: feature_id" or "Protein: feature_id"
            if ":" in content:
                feature_class, feature_id = content.split(":", 1)
                feature_meta[feature_class.strip()] = feature_id.strip() if feature_id.strip() else None
            else:
                feature_meta[content] = None

        # ---------------- colors (General) ----------------
        elif current_section == "colors":
            if "=" in content:
                key, value = content.split("=", 1)
                if key.strip() in colors_meta["global"]:
                    colors_meta["global"][key.strip()] = value.strip()

        # ---------------- colors.transcripts ----------------
        elif current_section == "colors.transcripts":
            if "=" in content:
                key, value = content.split("=", 1)
                colors_meta["transcripts"][key.strip()] = value.strip()

        # ---------------- colors.exon / cds ----------------
        elif current_section == "colors.exon":
            if "=" in content:
                key, value = content.split("=", 1)
                if key.strip() == "color": colors_meta["exon"]["color"] = value.strip()
        
        elif current_section == "colors.cds":
            if "=" in content:
                key, value = content.split("=", 1)
                if key.strip() == "color": colors_meta["cds"]["color"] = value.strip()

        # ---------------- colors.ff.* ----------------
        elif current_section == "colors.ff.continuous":
            if "=" in content:
                key, value = content.split("=", 1)
                if key.strip() == "cmap": colors_meta["ff"]["continuous"]["cmap"] = value.strip()

        elif current_section == "colors.ff.categorical":
            if "=" in content:
                key, value = content.split("=", 1)
                key, value = key.strip(), value.strip()
                if key == "palette":
                    colors_meta["ff"]["categorical"]["palette"] = value
                else:
                    # Specific category colors
                    colors_meta["ff"]["categorical"]["colors"][key] = value

        elif current_section == "colors.ff.binary":
            if "=" in content:
                key, value = content.split("=", 1)
                key, value = key.strip(), value.strip()
                if key in ("0", "1"):
                    colors_meta["ff"]["binary"][key] = value

    # finalize query
    query["tx"][0] = gtf_meta.get("transcript_id")
    if query["tx"][1] == set():
        query["tx"][1] = None

    return sample_meta, config_meta, ctrl_trgt_table, gtf_meta, colors_meta, query, feature_meta


def get_scale_unit(data_range):
    """プロットの幅に合わせて適切なスケールバーの単位とラベルを計算"""
    # 幅の約15%〜20%くらいの長さを目指す
    target_size = data_range * 0.15
    
    # 桁数を計算
    order_of_magnitude = 10 ** int(np.floor(np.log10(target_size)))
    
    # キリの良い倍率 (1, 2, 5)
    multipliers = [1, 2, 5, 10]
    
    selected_size = order_of_magnitude
    for m in multipliers:
        if order_of_magnitude * m <= target_size:
            selected_size = order_of_magnitude * m
        else:
            break
            
    # ラベル作成
    if selected_size >= 1000:
        label = f"{int(selected_size/1000)} kb"
    else:
        label = f"{int(selected_size)} bp"
        
    return selected_size, label


# -----------------------------
# isoform model
# -----------------------------
def get_isoform_model(gtf_file, gtf_meta, tx_colors, query, colors_meta=None):
    """Build exon/CDS models. Also complete transcript colors."""
    transcripts = {}
    transcripts_CDS = {}

    target_gene = query["gene"][1]
    target_tx = query["tx"][1]

    with open(gtf_file) as gtf:
        for line in gtf:
            if line.startswith("#"):
                continue
            fields = line.strip().split("\t")
            if len(fields) < 9: continue
            chrom, source, feature, start, end, score, strand, frame, attributes = fields
            chrom = chrom.replace("chr", "")

            attr_dict = {
                match.group(1): match.group(2)
                for match in re.finditer(r'(\S+)\s+"([^"]+)"', attributes)
            }
            transcript_id = attr_dict.get(gtf_meta["transcript_id"])
            line_gene = attr_dict.get(query["gene"][0])
            line_tx = attr_dict.get(query["tx"][0])

            if line_gene != target_gene:
                continue
            if target_tx is not None and line_tx not in target_tx:
                continue

            if feature == gtf_meta["exon"]:
                if transcript_id not in transcripts:
                    transcripts[transcript_id] = [[], 1 if strand == "+" else -1, chrom]
                transcripts[transcript_id][0].append((int(start), int(end)))

            if feature == gtf_meta.get("cds", "CDS"):
                if transcript_id not in transcripts_CDS:
                    transcripts_CDS[transcript_id] = [[], 1 if strand == "+" else -1, chrom]
                transcripts_CDS[transcript_id][0].append((int(start), int(end)))

    for tx in transcripts:
        transcripts[tx][0] = sorted(transcripts[tx][0], key=lambda x: x[0])
    for tx in transcripts_CDS:
        transcripts_CDS[tx][0] = sorted(transcripts_CDS[tx][0], key=lambda x: x[0])

    # complete colors
    default_tx = "#B3C8CF"
    palette_name = None
    if colors_meta:
        default_tx = colors_meta["global"].get("default_tx", default_tx)
        palette_name = colors_meta["global"].get("palette", None)

    missing = [tx for tx in transcripts if tx not in tx_colors]
    if missing:
        if palette_name:
            pal = sns.color_palette(palette_name, len(missing))
            for tx, c in zip(missing, pal):
                tx_colors[tx] = mcolors.to_hex(c)
        else:
            for tx in missing:
                tx_colors[tx] = default_tx

    return transcripts, transcripts_CDS, tx_colors


# -----------------------------
# feature model
# -----------------------------
def get_feature_model(gtf_file, gtf_meta, query, feature_meta):
    """Read positional functional features (GTF-based)."""
    ff_d = dict()

    target_gene = query["gene"][1]
    target_tx = query["tx"][1]

    with open(gtf_file) as gtf:
        for line in gtf:
            if line.startswith("#"):
                continue
            fields = line.strip().split("\t")
            if len(fields) < 9: continue
            chrom, source, feature, start, end, score, strand, frame, attributes = fields

            if feature in feature_meta:
                attr_dict = {
                    match.group(1): match.group(2)
                    for match in re.finditer(r'(\S+)\s+"([^"]+)"', attributes)
                }
                transcript_id = attr_dict.get(gtf_meta["transcript_id"])

                line_gene = attr_dict.get(query["gene"][0])
                line_tx = attr_dict.get(query["tx"][0])
                if line_gene != target_gene:
                    continue
                if target_tx is not None and line_tx not in target_tx:
                    continue

                if transcript_id not in ff_d:
                    ff_d[transcript_id] = {feat: {} for feat in feature_meta}

                if feature_meta[feature] is None:
                    if ff_d[transcript_id][feature] == {}:
                        ff_d[transcript_id][feature]["NONAME"] = []
                    ff_d[transcript_id][feature]["NONAME"].append((int(start), int(end)))
                else:
                    feat_id = attr_dict.get(feature_meta[feature])
                    if feat_id:
                        if feat_id not in ff_d[transcript_id][feature]:
                            ff_d[transcript_id][feature][feat_id] = []
                        ff_d[transcript_id][feature][feat_id].append((int(start), int(end)))

    for tx in ff_d:
        for feat in feature_meta:
            for ind in ff_d[tx][feat]:
                ff_d[tx][feat][ind] = sorted(ff_d[tx][feat][ind], key=lambda x: x[0])

    return ff_d


def formatting_isoform_model(transcripts_data, transcripts, annot):
    if annot == "exon":
        for transcript_id, exons in transcripts.items():
            if not exons[0]: continue
            start = min([i[0] for i in exons[0]])
            end = max(i[1] for i in exons[0])
            transcripts_data.append(
                {
                    "id": transcript_id,
                    "exons": exons[0],
                    "strand": exons[1],
                    "seq_region_name": exons[2],
                    "start": start,
                    "end": end,
                }
            )
    elif annot == "cds":
        for i in range(len(transcripts_data)):
            isomodel = transcripts_data[i]
            tx_id = isomodel["id"]
            if tx_id in transcripts:
                transcripts_data[i]["cds"] = transcripts[tx_id][0]
            else:
                transcripts_data[i]["cds"] = []
    return transcripts_data


# -----------------------------
# nonGTF annotations
# -----------------------------
def get_nonGTF_annotations(annotation_file, gene):
    if annotation_file is None:
        return {}

    nongtf_d = dict()
    with open(annotation_file) as f:
        for line in f:
            line = line.strip()
            if line.startswith("#") or line == "":
                continue
            cols = line.split("\t")
            if len(cols) < 5: continue
            tx_id, gene_id, feat_name, feat_type, status = cols
            if gene != gene_id:
                continue
            if tx_id not in nongtf_d:
                nongtf_d[tx_id] = {}
            nongtf_d[tx_id][feat_name] = [feat_type, status]

    features_d = dict()
    for tx_id in nongtf_d:
        for feat in nongtf_d[tx_id]:
            feat_type = nongtf_d[tx_id][feat][0]
            if feat not in features_d:
                features_d[feat] = [feat_type, {}]
    for feat in features_d:
        features_d[feat][1] = {tx_id: None for tx_id in nongtf_d}
    for tx_id in nongtf_d:
        for feat in nongtf_d[tx_id]:
            status = nongtf_d[tx_id][feat][1]
            features_d[feat][1][tx_id] = status

    return features_d


def prepare_ax1_xaxis(ax1, ci, x_min, x_max, x_min_eff, x_max_eff):
    if ci is None:
        ax1.xaxis.set_major_formatter(ticker.FormatStrFormatter("%d"))
    else:
        ax1.set_xticks([x_min_eff, x_max_eff])
        ax1.set_xticklabels([str(x_min), str(x_max)])
    return ax1


def coordinate_adjustment(feat_names, Y_LOW, Y_UPP):
    out_d = dict()
    if feat_names == []:
        return out_d
    width = (Y_UPP - Y_LOW) / len(feat_names)
    for i in range(len(feat_names)):
        out_d[feat_names[i]] = (Y_UPP - width * (i + 1), Y_UPP - width * i)
    return out_d


def darken_color(color, factor=0.4):
    return tuple(max(0, c * factor) for c in color)


def overlaped_feature_indivisuals(DICT):
    def overlaps(r1, r2):
        return r1[0] <= r2[1] and r2[0] <= r1[1]

    graph = defaultdict(set)
    keys = list(DICT.keys())

    for i in range(len(keys)):
        for j in range(i + 1, len(keys)):
            k1, k2 = keys[i], keys[j]
            for r1 in DICT[k1]:
                for r2 in DICT[k2]:
                    if overlaps(r1, r2):
                        graph[k1].add(k2)
                        graph[k2].add(k1)
                        break
                else:
                    continue
                break

    visited = set()
    groups = []

    def dfs(node, group):
        visited.add(node)
        group.append(node)
        for neighbor in graph[node]:
            if neighbor not in visited:
                dfs(neighbor, group)

    for k in keys:
        if k not in visited:
            group = []
            dfs(k, group)
            groups.append(group)

    return groups


# -----------------------------
# ax1 drawing
# -----------------------------
def prepare_ff_ax1(transcripts_data, model_features, ax1, gene_name, tx_colors, colors_meta, tss_mode, ci=None, hide_scale=False):
    if not transcripts_data: return ax1
    
    MIN = min(transcripts_data, key=lambda x: x["start"])["start"]
    MAX = max(transcripts_data, key=lambda x: x["end"])["end"]
    # Reverse order for plotting (Top=First in list)
    transcripts_data = transcripts_data[::-1]

    y_positions = []
    
    if model_features == {}:
        feat_names = []
        feat_colors = []
    else:
        # Check first existing transcript in model_features to get keys
        first_key = next(iter(model_features), None)
        if first_key:
            feat_names = list(model_features[first_key].keys())
            feat_colors = sns.color_palette("husl", len(feat_names))
        else:
            feat_names = []
            feat_colors = []

    all_tss = []
    
    # global colors
    gcol = colors_meta.get("global", {})
    line_color = gcol.get("default_line", "gray")
    text_color = gcol.get("default_text", "black")
    default_tx_color = gcol.get("default_tx", "#B3C8CF")

    user_tx_colors = colors_meta.get("user_transcripts", {})
    exon_override = colors_meta.get("exon", {}).get("color", None)
    cds_override = colors_meta.get("cds", {}).get("color", None)
    default_tx_color = colors_meta.get("global", {}).get("default_tx", "#999999")
    default_cds_color = colors_meta.get("global", {}).get("default_cds", "#3b73b9")

    for i, transcript_data in enumerate(transcripts_data):
        y_positions.append(i)
        start = transcript_data["start"]
        end = transcript_data["end"]
        tx_name = transcript_data["id"]

        strand = transcript_data["strand"]
        arrow_direction = "right" if strand == 1 else "left"
        y_pos = i

        if strand == 1:
            all_tss.append(start)
        else:
            all_tss.append(end)

        interval = max(1, (MAX - MIN) // 50)
        x_positions = np.arange(start, end, interval)[1:-1]

        ax1.annotate(
            "",
            xy=(end, y_pos),
            xytext=(start, y_pos),
            arrowprops=dict(arrowstyle="-", color=line_color, lw=1),
        )

        if arrow_direction == "right":
            ax1.scatter(x_positions, [y_pos] * len(x_positions), marker=">", color=line_color, s=10)
        else:
            ax1.scatter(x_positions, [y_pos] * len(x_positions), marker="<", color=line_color, s=10)

        
        # 1. Exon色の決定
        if tx_name in user_tx_colors:
            current_exon_color = user_tx_colors[tx_name]
        elif exon_override:
            current_exon_color = exon_override
        else:
            current_exon_color = tx_colors.get(tx_name, default_tx_color)

        # 2. CDS色の決定
        base_color = tx_colors.get(tx_name, default_tx_color)
        if cds_override:
            current_cds_color = cds_override
        elif tx_name in user_tx_colors:
            current_cds_color = user_tx_colors[tx_name]
        elif base_color == default_tx_color:
            current_cds_color = default_cds_color
        else:
            current_cds_color = base_color

        # exon
        for exon in transcript_data["exons"]:
            exon_start, exon_end = exon
            ax1.add_patch(
                patches.Rectangle(
                    (exon_start, i - 0.04),
                    exon_end - exon_start,
                    0.08,
                    color=current_exon_color,
                )
            )

        # CDS
        for cds in transcript_data["cds"]:
            cds_start, cds_end = cds
            ax1.add_patch(
                patches.Rectangle(
                    (cds_start, i - 0.08),
                    cds_end - cds_start,
                    0.16,
                    color=current_cds_color,
                )
            )

        # positional features
        if tx_name not in model_features:
            continue

        Y_LOW = i - 0.8
        Y_UPP = i - 0.11
        coord_d = coordinate_adjustment(feat_names, Y_LOW, Y_UPP)

        for k_feat, feat in enumerate(feat_names):
            if feat not in model_features[tx_name]: continue
            
            y_low, y_upp = coord_d[feat]
            plot_order = sorted(
                model_features[tx_name][feat],
                key=lambda k: sum(e - s for s, e in model_features[tx_name][feat][k]),
                reverse=True,
            )
            plot_groups = overlaped_feature_indivisuals(model_features[tx_name][feat])
            plot_groups_cnt = [-1 for _ in range(len(plot_groups))]

            for ind in plot_order:
                idx = 0
                for g in range(len(plot_groups)):
                    if ind in plot_groups[g]:
                        plot_groups_cnt[g] += 1
                        idx = g
                        break

                y_low2 = y_low - 0.009 * plot_groups_cnt[idx]
                y_upp2 = y_upp - 0.009 * plot_groups_cnt[idx]
                crush_rate = 0.83

                for (start_k, end_k) in model_features[tx_name][feat][ind]:
                    ax1.add_patch(
                        patches.Rectangle(
                            (start_k, y_upp2 - (y_upp2 - y_low2) * crush_rate),
                            end_k - start_k,
                            (y_upp2 - y_low2) * crush_rate,
                            color=feat_colors[k_feat],
                        )
                    )

                leftmost_x = model_features[tx_name][feat][ind][0][0]
                rightmost_x = model_features[tx_name][feat][ind][-1][1]
                ax1.add_patch(
                    patches.Rectangle(
                        (leftmost_x, y_upp2 - (y_upp2 - y_low2) * crush_rate),
                        rightmost_x - leftmost_x,
                        (y_upp2 - y_low2) * crush_rate,
                        linewidth=0.5,
                        edgecolor=darken_color(feat_colors[k_feat]),
                        facecolor="none",
                    )
                )

    if tss_mode:
        unique_tss = sorted(set(all_tss))
        ymin = -0.3
        ymax = len(transcripts_data) - 0.8
        for x in unique_tss:
            ax1.vlines(
                x,
                ymin=ymin,
                ymax=ymax,
                colors=line_color,
                linestyles=":",
                linewidth=0.8,
                alpha=0.4,
            )

    # === スケールバーの描画 ===
    if not hide_scale:
        data_width = MAX - MIN
        scale_size, scale_label = get_scale_unit(data_width)
        bar_y = - 0.8
        text_y = bar_y - 0.05
        bar_end = MAX
        bar_start = MAX - scale_size
        ax1.plot([bar_start, bar_end], [bar_y, bar_y], color="#3B5894", linewidth=1, clip_on=False)
        cap_height = 0.1
        ax1.plot([bar_start, bar_start], [bar_y - cap_height/2, bar_y + cap_height/2], color="#3B5894", linewidth=1, clip_on=False)
        ax1.plot([bar_end, bar_end], [bar_y - cap_height/2, bar_y + cap_height/2], color="#3B5894", linewidth=1, clip_on=False)
        ax1.text((bar_start + bar_end) / 2, text_y, scale_label, ha='center', va='top', fontsize=8, color="#3B5894")
        if ci is not None:
            note_text = f"Note: Introns compressed"
            ax1.text(bar_end, bar_y + 0.1, note_text, ha='right', va='bottom', fontsize=8, color="#3B5894", style='italic')
    # ===========================

    space = int((MAX - MIN) / 20)
    ax1.set_xlim(MIN - 2 * space, MAX + space)
    ax1.set_ylim(-1.0, len(transcripts_data) - 0.5)
    
    chrom_label = transcripts_data[0]["seq_region_name"]
    chrom_norm = chrom_label.lower()
    if chrom_norm.startswith("chr"):
        chrom_norm = chrom_label[3:]
    else:
        chrom_norm = chrom_label

    ax1.set_xlabel(f'Chr{chrom_norm}', color=text_color)
    ax1.set_title(gene_name, color=text_color)
    ax1.xaxis.set_major_formatter(ticker.FormatStrFormatter("%d"))
    ax1.set_yticks(y_positions)
    transcripts = [i["id"] for i in transcripts_data]
    ax1.set_yticklabels(transcripts, color=text_color)

    # feature legend labels
    N = len(transcripts_data) - 1
    if feat_names:
        coord = coordinate_adjustment(feat_names, Y_LOW=N - 0.8, Y_UPP=N - 0.11)
        for k_feat, feat in enumerate(feat_names):
            y_low, y_upp = coord[feat]
            ax1.text(
                MIN - 1.8 * space,
                (y_upp + y_low) / 2,
                feat,
                fontsize=9,
                color=darken_color(feat_colors[k_feat], factor=0.6),
                ha="left",
                va="center",
            )

    for label in ax1.get_xticklabels():
        label.set_rotation(15)
        label.set_ha("right")
        label.set_color(text_color)

    return ax1


def draw_combined_legend(ax, legend_metadata):
    ax.axis("off")
    n = len(legend_metadata)
    if n == 0: return

    slot_height = 1.0 / n
    
    for i, meta in enumerate(legend_metadata):
        y = 1.0 - (i + 1) * slot_height
        
        # 凡例用の小領域を作成
        sub_ax = ax.inset_axes([0.1, y + 0.05 * slot_height, 0.9, slot_height * 0.9], transform=ax.transAxes)
        sub_ax.axis("off")
        
        sub_ax.set_title(meta["title"], loc="left", fontsize=10, fontweight='bold')
        
        mode = meta["mode"]
        info = meta["info"]
        
        if mode == "discrete" and info:
            # カテゴリカル / バイナリ
            # 変更点: facecolorとedgecolorを分けて指定し、linewidthを追加
            handles = [
                patches.Patch(
                    facecolor=c, 
                    edgecolor="black", 
                    linewidth=1.0, 
                    label=str(l)
                ) for l, c in info.items()
            ]
            
            sub_ax.legend(
                handles=handles, 
                loc='upper left', 
                bbox_to_anchor=(0, 1.0), 
                fontsize=9, 
                frameon=False,
                borderaxespad=0
            )
            
        elif mode == "continuous" and info:
            cbar_ax = sub_ax.inset_axes([0, 0.6, 0.6, 0.15], transform=sub_ax.transAxes)
            
            sm = ScalarMappable(cmap=info["cmap"], norm=info["norm"])
            sm.set_array([])
            
            cb = plt.colorbar(sm, cax=cbar_ax, orientation='horizontal')
            cbar_ax.xaxis.set_ticks_position('bottom')
            cbar_ax.tick_params(labelsize=8)
            
            cb.outline.set_edgecolor('black')
            cb.outline.set_linewidth(1)
            
            cbar_ax.set_xticks([info["norm"].vmin, info["norm"].vmax])
            cbar_ax.set_xticklabels([f"{info['norm'].vmin:.1f}", f"{info['norm'].vmax:.1f}"])


# axk drawing
def prepare_ff_axk(ax_l, transcripts_data, annotation_dataB, ff_colors_meta):
    def is_float(v):
        try:
            float(v)
            return True
        except (ValueError, TypeError):
            return False

    def Color_binary(A):
        bmap = ff_colors_meta.get("binary", {})
        c0 = mcolors.to_rgba(bmap.get("0", "#cccccc"))
        c1 = mcolors.to_rgba(bmap.get("1", "#000000"))
        legend_map = {"0": c0, "1": c1}
        binary_colors = {"0": c0, "1": c1, None: "white"}
        mapping = {key: binary_colors.get(str(value), "white") if value is not None else "white" for key, value in A.items()}
        return mapping, legend_map, "discrete"

    def Color_categorical(A):
        unique_categories = sorted(set(v for v in A.values() if v is not None))
        cat_meta = ff_colors_meta.get("categorical", {})
        manual = cat_meta.get("colors", {})
        palette_name = cat_meta.get("palette", "tab10")
        
        category_colors = {cat: manual[cat] for cat in manual if cat in unique_categories}
        remaining = [cat for cat in unique_categories if cat not in category_colors]
        
        if remaining:
            auto_cols = None
            # ... (自動色生成ロジック省略: 変更なし) ...
            try:
                cmap = plt.get_cmap(palette_name, len(remaining))
                auto_cols = [cmap(i) for i in range(len(remaining))]
            except ValueError:
                pal = sns.color_palette("tab10", len(remaining))
                auto_cols = [mcolors.to_rgba(c) for c in pal]
            for cat, col in zip(remaining, auto_cols):
                category_colors[cat] = col

        legend_map = category_colors.copy()
        category_colors[None] = "white"
        mapping = {key: category_colors[value] for key, value in A.items()}
        return mapping, legend_map, "discrete"

    def Color_continuous(A):
        valid_values = [float(v) for v in A.values() if v is not None and is_float(v)]
        if not valid_values:
            return {key: "white" for key in A}, None, "continuous"
        vmin, vmax = min(valid_values), max(valid_values)
        if vmin == vmax: vmax = vmin + 1
        cmap_name = ff_colors_meta.get("continuous", {}).get("cmap", "viridis")
        cmap = plt.get_cmap(cmap_name)
        norm = mcolors.Normalize(vmin=vmin, vmax=vmax)
        mapping = {key: cmap(norm(float(value))) if value is not None and is_float(value) else "white" for key, value in A.items()}
        legend_info = {"cmap": cmap, "norm": norm}
        return mapping, legend_info, "continuous"

    # --- メイン処理 ---
    # 凡例メタデータを収集するためのリスト
    legend_metadata = []
    features_l = list(annotation_dataB.keys())

    # 各特徴量ごとに色情報を計算してメタデータを保存
    for feat in features_l:
        CATEGORY = annotation_dataB[feat][0]
        VALUES = annotation_dataB[feat][1]
        
        if CATEGORY == "binary":
            _, info, mode = Color_binary(VALUES)
        elif CATEGORY == "categorical":
            _, info, mode = Color_categorical(VALUES)
        elif CATEGORY == "continuous":
            _, info, mode = Color_continuous(VALUES)
        else:
            info, mode = None, None
        
        legend_metadata.append({"title": feat, "mode": mode, "info": info})

    # 描画ループ (凡例描画部分は削除)
    N = len(transcripts_data)
    for i, transcript_data in enumerate(transcripts_data):
        tx_name = transcript_data["id"]
        for j, feat in enumerate(features_l):
            CATEGORY = annotation_dataB[feat][0]
            VALUES = annotation_dataB[feat][1]
            
            if CATEGORY == "binary": mapping, _, _ = Color_binary(VALUES)
            elif CATEGORY == "categorical": mapping, _, _ = Color_categorical(VALUES)
            elif CATEGORY == "continuous": mapping, _, _ = Color_continuous(VALUES)
            else: mapping = {k: "white" for k in VALUES}

            hatch = "//" if VALUES.get(tx_name) is None else None
            rect = patches.Rectangle(
                (0.3, N - i - 0.9 - 1), 0.4, 1,
                facecolor=mapping.get(tx_name, "white"),
                edgecolor="black", hatch=hatch, linewidth=2
            )
            ax_l[j].add_patch(rect)
            
            val_disp = VALUES.get(tx_name) if VALUES.get(tx_name) is not None else ""
            ax_l[j].text(0.5, N - i - 0.9 - 0.5, val_disp, ha="center", va="center", rotation="vertical", fontsize=10)

    for j, feat in enumerate(features_l):
        ax_l[j].set_title(feat)

    for ax in ax_l:
        ax.set_xlim(0, 1)
        ax.set_ylim(-1.0, N - 0.5)
        ax.axis("off")

    return ax_l, legend_metadata


def reorder(transcripts_data, meta_data):
    if "order" not in meta_data:
        return transcripts_data
    tmp = []
    order = meta_data["order"]
    for tx in order:
        for i in transcripts_data:
            if i["id"] == tx:
                tmp.append(i)
                break
    return tmp


def plot_isoespy_ff(transcripts_data, model_features, config_meta, gene_name,
                    meta_data, tx_colors, colors_meta, ci,
                    x_min, x_max, x_min_eff, x_max_eff, annotation_dataB, tss_mode,
                    output_file=None, dpi=300, hide_scale=False):
    transcripts_data = reorder(transcripts_data, meta_data)
    if not transcripts_data:
        print("No transcripts found to plot.")
        return

    N = len(annotation_dataB.keys())
    
    # レイアウト変更: Main + N個のヒートマップ + 1個の凡例スペース
    # width_ratios: メイン(20), ヒートマップ(1...1), 凡例(4)
    total_cols = 1 + N + (1 if N > 0 else 0) # ヒートマップがない場合は凡例もなし
    
    if N == 0:
        fig, ax1 = plt.subplots(ncols=1, figsize=(20, 8))
        ax_l = []
        ax_legend = None
    else:
        # 凡例用スペースを最後に追加
        ratios = [20] + [1] * N + [4]
        fig, axes = plt.subplots(
            ncols=len(ratios),
            figsize=(20 + N + 2, 8), # 幅を少し広げる
            gridspec_kw={"width_ratios": ratios},
        )
        ax1 = axes[0]
        ax_l = axes[1:-1] # 中間がヒートマップ
        ax_legend = axes[-1] # 最後が凡例

    ax1 = prepare_ff_ax1(transcripts_data, model_features, ax1, gene_name, tx_colors, colors_meta, tss_mode, ci, hide_scale)
    ax1 = prepare_ax1_xaxis(ax1, ci, x_min, x_max, x_min_eff, x_max_eff)

    if N != 0:
        # prepare_ff_axk が legend_metadata も返すようになったので受け取る
        ax_l, legend_metadata = prepare_ff_axk(ax_l, transcripts_data, annotation_dataB, colors_meta["ff"])
        
        # まとめて凡例を描画
        draw_combined_legend(ax_legend, legend_metadata)

    plt.subplots_adjust(wspace=0.1) # 間隔調整

    if output_file:
        try:
            fig.savefig(output_file, dpi=dpi, bbox_inches='tight')
            print(f"Figure successfully saved to {output_file} with DPI={dpi}")
        except Exception as e:
            print(f"Error saving figure to file: {e}")
            plt.show()
    else:
        plt.show()


def moved_data_for_exons_cds(main_data, ci):
    model = {}
    for tx_data in main_data:
        tx_name = tx_data["id"]
        model[tx_name + "_exons"] = tx_data["exons"]
        model[tx_name + "_cds"] = tx_data["cds"]

    model_compressed = intronCompression(model, ci)

    startend_d = {}
    for tx_data in main_data:
        tx_name = tx_data["id"]
        if tx_name + "_exons" in model_compressed and model_compressed[tx_name + "_exons"]:
            start = model_compressed[tx_name + "_exons"][0][0]
            end = model_compressed[tx_name + "_exons"][-1][1]
            startend_d[tx_name] = {"start": start, "end": end}
        else:
             startend_d[tx_name] = {"start": tx_data['start'], "end": tx_data['end']}

    for tx_data in main_data:
        tx_name = tx_data["id"]
        tx_data["exons"] = model_compressed[tx_name + "_exons"]
        tx_data["cds"] = model_compressed[tx_name + "_cds"]
        tx_data["start"] = startend_d[tx_name]["start"]
        tx_data["end"] = startend_d[tx_name]["end"]

    return main_data


def moved_data_for_features(model_features, main_data, ci):
    model = {}
    # ベースのExon/CDSを登録
    for tx_data in main_data:
        tx_name = tx_data["id"]
        model[tx_name + "_exons"] = tx_data["exons"]
        model[tx_name + "_cds"] = tx_data["cds"]

    # 特徴量の座標を登録
    for tx_name in model_features:
        for feat_name in model_features[tx_name]:
            for id_name in model_features[tx_name][feat_name]:
                model[tx_name + "_" + feat_name + "_" + id_name] = model_features[tx_name][feat_name][id_name]

    model_compressed = intronCompression(model, ci)

    # 変換後の座標を戻す
    for tx_name in model_features:
        for feat_name in model_features[tx_name]:
            for id_name in model_features[tx_name][feat_name]:
                key = tx_name + "_" + feat_name + "_" + id_name
                if key in model_compressed:
                    model_features[tx_name][feat_name][id_name] = model_compressed[key]

    return model_features


def isoespy_ff(gene, gtf_data, meta_data, ci, annotation_file, tss_mode, output_file=None, dpi=300, hide_scale=False):
    sample_meta, config_meta, ctrl_trgt_table, gtf_meta, colors_meta, query, feature_meta = parse_metadata(meta_data, gene)

    # isoform model + colors completion
    tx_colors = colors_meta["transcripts"]
    model_exon, model_cds, tx_colors = get_isoform_model(gtf_data, gtf_meta, tx_colors, query, colors_meta)

    # feature model (Genomic Feature)
    model_features = get_feature_model(gtf_data, gtf_meta, query, feature_meta)

    # formatted isoform model
    transcripts_main_data = []
    transcripts_main_data = formatting_isoform_model(transcripts_main_data, model_exon, annot="exon")
    transcripts_main_data = formatting_isoform_model(transcripts_main_data, model_cds, annot="cds")
    
    if not transcripts_main_data:
        print(f"No transcripts found for gene: {gene}")
        return

    x_min = min(transcripts_main_data, key=lambda x: x["start"])["start"]
    x_max = max(transcripts_main_data, key=lambda x: x["end"])["end"]

    if ci is not None:
        copied_data = copy.deepcopy(transcripts_main_data)
        transcripts_main_data = moved_data_for_exons_cds(transcripts_main_data, ci)
        # copyした元データを使って feature も圧縮対応
        model_features = moved_data_for_features(model_features, copied_data, ci)

    x_min_eff = min(transcripts_main_data, key=lambda x: x["start"])["start"]
    x_max_eff = max(transcripts_main_data, key=lambda x: x["end"])["end"]

    # Functional annotation data (heatmap)
    annotation_dataB = get_nonGTF_annotations(annotation_file, gene)

    plot_isoespy_ff(
        transcripts_main_data,
        model_features,
        config_meta,
        gene,
        config_meta,
        tx_colors,
        colors_meta,
        ci,
        x_min,
        x_max,
        x_min_eff,
        x_max_eff,
        annotation_dataB,
        tss_mode,
        output_file,
        dpi,
        hide_scale
    )


def process_ci(ci):
    if ci is None:
        return None
    if isinstance(ci, str):
        try:
            ci = float(ci)
        except ValueError:
            warnings.warn(f"Warning: ci should be a float/int, but received string '{ci}' that cannot be converted.")
            return None
    else:
        try:
            ci = float(ci)
        except (ValueError, TypeError):
            warnings.warn(f"Warning: ci should be a float/int, but received '{ci}' that cannot be converted.")
            return None
    return ci


def main(args=None):
    parser = argparse.ArgumentParser(description="isoespy_ff()")
    parser.add_argument("-gene", "--gene_name", required=True, type=str, default=None, help="Gene name")
    parser.add_argument("-gtf", "--gtf_data", required=True, type=str, default=None, help="GTF file")
    parser.add_argument("-meta", "--meta_data", required=True, type=str, default=None, help="metadata")
    parser.add_argument("-ci", "--compress_introns", default=None, help="intron compression parameter")
    parser.add_argument("-a", "--annotation", type=str, default=None, help="non-gtf transcript annotation file")
    parser.add_argument("-tss", "--tss_line", action="store_true", help="Show TSS support lines")
    parser.add_argument("-o", "--output_file", type=str, default=None, help="Save figure to file (e.g., plot.pdf or plot.png)")
    parser.add_argument("--dpi", type=int, default=300, help="Resolution for image formats (e.g., PNG) in DPI")
    parser.add_argument("--hide_scale", action="store_true", help="Hide scale bar")

    args = parser.parse_args()
    gene = args.gene_name
    gtf_data = args.gtf_data
    meta_data = args.meta_data
    ci = process_ci(args.compress_introns)
    annotation_data = args.annotation
    tss_mode = args.tss_line
    output_file = args.output_file
    dpi = args.dpi
    hide_scale = args.hide_scale

    isoespy_ff(gene, gtf_data, meta_data, ci, annotation_data, tss_mode, output_file, dpi, hide_scale)


if __name__ == "__main__":
    main()



""" 転写産物上のアミノ酸範囲 ==> 転写産物上の塩基配列範囲 ==> ゲノムエクソン上の範囲
　　に変換するモジュール
"""

# ================= Input =================
# arg1 - GTF (CDS feature 必須)
# arg2 - 転写産物名
# arg3 - アミノ酸範囲 (1-based, inclusive)
# arg4 - cds全長

# ================= Output =================
# ゲノムにおける範囲 (list)　e.g., [(start1, end1), (start2, end2), (start3, end3)]    **start昇順


def aanucConverter(aafrom, aato, frame):
    """ アミノ酸配列上範囲 ==> CDS座標上範囲 """
    nucfrom = 3*aafrom - 2 + frame
    nucto = 3*aato + frame
    return (nucfrom, nucto)

def map_feature_to_genome(coord_tx, coord_gen, strand_gen):
    """ 塩基配列上範囲 ==> ゲノム上範囲 """

    start_tx = coord_tx["feature_start"]	# 転写産物塩基配列上のfeature 開始点 CDS (aa), exon (nuc)
    end_tx = coord_tx["feature_end"]		# 転写産物塩基配列上のfeature 終了点
    totallen_tx = coord_tx["total_len"]		# CDS全長 (aa) / exon全長 (nuc)

    if strand_gen == "+":   # ゲノム上でアイソフォームが+鎖にコード
        start_tx, end_tx = start_tx, end_tx
    else:                   # ゲノム上でアイソフォームが-鎖にコード
        start_tx, end_tx = totallen_tx - end_tx + 1, totallen_tx - start_tx + 1

    exon_table = {"tx": [], "gen": []}		# tx: 転写産物上位置　　gen: ゲノム上位置
    feature_table = {"gen": []}			# ゲノム上 feature 位置

    # exon 位置対応表 tx --- genome を作成
    exon_table["gen"] = coord_gen
    cds_pos = 0
    for exon_i in exon_table["gen"]:
        exon_len = exon_i[1] - exon_i[0] + 1
        exon_table["tx"].append((cds_pos+1, cds_pos+exon_len))
        cds_pos += exon_len

    #print("exon_table", exon_table)
    if totallen_tx != exon_table["tx"][-1][1]:
        print("この転写産物は全長が不一致です!!!")

    #print(exon_table)

    # feature 座標をゲノムにマップする
    for i in range(len(exon_table["tx"])):
        if exon_table["tx"][i][0] <= start_tx <= exon_table["tx"][i][1]:
            start_exon = i + 1									# start_exon: feature が開始するエクソン番号 (参照ゲノム+鎖上流から数えて)
            start_pos = exon_table["gen"][i][0] + (start_tx - exon_table["tx"][i][0])		# start_pos: ゲノム上のfeature 開始点
        if exon_table["tx"][i][0] <= end_tx <= exon_table["tx"][i][1]:
            end_exon = i + 1									# end_exon: feature が終了するエクソン番号 (参照ゲノム+鎖上流から数えて)
            end_pos = exon_table["gen"][i][0] + (end_tx - exon_table["tx"][i][0])		# end_pos: ゲノム上のfeature 終了点

    if start_exon == end_exon:   # feature が１つのエクソン内で完結
        feature_table["gen"].append((start_pos, end_pos))
    else:                        # feature が複数のエクソンにまたがる
        for i in range(start_exon-1, end_exon):
            if i == start_exon-1:   # CDSを含む始めのエクソン
                feature_table["gen"].append((start_pos, exon_table["gen"][i][1]))
            elif i == end_exon-1:   # CDSを含む最後のエクソン
                feature_table["gen"].append((exon_table["gen"][i][0], end_pos))
            else:                   # CDSを含む内部のエクソン
                feature_table["gen"].append(exon_table["gen"][i])

    #print("feature_table")
    #print(feature_table)
    return feature_table["gen"]   # タプルリストを返す


def backmap_aa(coord_gen, tx, aafrom, aato, cds_total_len, strand_gen, frame_d):
    """ coord_gen : "CDS" のゲノム座標上での位置情報 list形式
        tx        : 転写産物名
        aafrom    : アミノ酸配列上におけるfeature (ドメインなど) の開始　  1-based, inclusive
        aato      : アミノ酸配列上におけるfeature (ドメインなど) の終了　　1-based, inclusive
    """
    if tx in frame_d:
        frame = frame_d[tx]
    else:
        # フレームデータが無い時は frame = 0 と仮定
        frame = 0

    # アミノ酸配列上位置 ==> 塩基配列上配列位置 (CDS座標上, 転写産物座標上ではない)
    nucfrom, nucto = aanucConverter(aafrom, aato, frame)

    # CDS座標上位置 ==> ゲノム上位置
    coord_cds = {"feature_start": nucfrom, "feature_end": nucto, "total_len": cds_total_len}
    backmap_coords = map_feature_to_genome(coord_cds, coord_gen, strand_gen)
    return backmap_coords


def backmap_nuc(coord_gen, tx, nucfrom, nucto, exon_total_len, strand_gen):
    """ coord_gen : "exon" のゲノム座標上での位置情報 list形式
        tx        : 転写産物名
        nucfrom   : 塩基配列上におけるfeature (CDSなど) の開始　  1-based, inclusive
        nucto     : 塩基配列上におけるfeature (CDSなど) の終了　　1-based, inclusive
    """
    # 転写産物上位置 ==> ゲノム上位置
    coord_exon = {"feature_start": nucfrom, "feature_end": nucto, "total_len": exon_total_len}
    backmap_coords = map_feature_to_genome(coord_exon, coord_gen, strand_gen)
    return backmap_coords


if __name__ == "__main__":
    print("Call from another script.")




import argparse
import csv
import sys
import re
from collections import OrderedDict

def parse_args():
    p = argparse.ArgumentParser(
        description="Create DRIMSeq input table (gene_id, feature_id, counts...) from TSV-GTF and counts matrix."
    )
    p.add_argument("--gtf-tsv", required=True, help="GTF as TSV (attributes in 9th column)")
    p.add_argument("--counts", required=True, help="Counts matrix TSV (col1=tx, col2+=samples)")
    p.add_argument("--gene-key", default="gene_symbol", help="Attribute key for gene id (default: gene_symbol)")
    p.add_argument("--tx-key", default="transcript_id", help="Attribute key for transcript id (default: transcript_id)")
    p.add_argument("--output", default="drimseq_counts.tsv", help="Output TSV (default: drimseq_counts.tsv)")
    return p.parse_args()

def parse_attributes(attr_field):
    """
      - key "value"; key "value";
      - key=value;key=value;
      - key value; のような形式
    """
    attr = {}
    if attr_field is None:
        return attr
    s = attr_field.strip().strip(";")
    if not s:
        return attr

    # ; で分割（連続セミコロンや末尾空要素を無視）
    parts = re.split(r';\s*', s)
    for part in parts:
        part = part.strip()
        if not part:
            continue
        key = None
        val = None

        if '=' in part:
            key, val = part.split('=', 1)
        else:
            # key "value" あるいは key value を拾う
            m = re.match(r'(\S+)\s+"?(.*?)"?$', part)
            if m:
                key, val = m.group(1), m.group(2)

        if key is not None:
            key = key.strip()
            if val is not None:
                val = val.strip().strip('"')
            attr[key] = val
    return attr

def build_tx_to_gene_map(gtf_tsv_path, gene_key, tx_key):
    """
    第9列 attributes から gene_key と tx_key を取り出し、
    転写産物ID -> 遺伝子ID の辞書を作る。
    同一転写産物IDが複数行に出ても最初の対応を採用（2回目以降は無視）。
    """
    tx2gene = OrderedDict()
    line_no = 0
    with open(gtf_tsv_path, "r", encoding="utf-8-sig", newline="") as f:
        reader = csv.reader(f, delimiter="\t")
        for row in reader:
            line_no += 1
            if not row or len(row) < 9:
                # 不完全行はスキップ
                continue
            attributes_field = row[8]
            attrs = parse_attributes(attributes_field)

            tx = attrs.get(tx_key)
            gene = attrs.get(gene_key)

            if tx is None or gene is None:
                # 必要キーが無い行はスキップ（構造上、exon行など全てに両方あるとは限らないため）
                continue

            if tx not in tx2gene:
                tx2gene[tx] = gene
            else:
                # 2回目以降は飛ばす（ただし gene が不一致なら警告）
                if tx2gene[tx] != gene:
                    print(
                        f"[WARN] Transcript {tx!r} already mapped to {tx2gene[tx]!r}, "
                        f"but found another gene {gene!r} at line {line_no}. Keeping the first mapping.",
                        file=sys.stderr
                    )
    if not tx2gene:
        print("[ERROR] No (transcript_id, gene_id) pairs were extracted from attributes. "
              f"Check that 9th column contains attributes and keys '{tx_key}', '{gene_key}'.",
              file=sys.stderr)
        sys.exit(1)
    return tx2gene

def read_counts(counts_path):
    """
    counts TSV を読み込み：
      - 1行目: ヘッダー（col1=tx、col2+=サンプル名）
      - 2行目以降: tx, sample1_count, sample2_count, ...
    戻り値:
      header_samples: [sample1, sample2, ...]
      rows: [(tx, [c1, c2, ...]), ...]  ※入力の順序維持
    """
    rows = []
    with open(counts_path, "r", encoding="utf-8-sig", newline="") as f:
        reader = csv.reader(f, delimiter="\t")
        try:
            header = next(reader)
        except StopIteration:
            print("[ERROR] Counts file is empty.", file=sys.stderr)
            sys.exit(1)
        if len(header) < 2:
            print("[ERROR] Counts header must have at least two columns: tx and >=1 sample.", file=sys.stderr)
            sys.exit(1)

        # 1列目は tx、2列目以降はサンプル名
        sample_headers = header[1:]

        for i, row in enumerate(reader, start=2):
            if not row:
                continue
            if len(row) < 2:
                print(f"[ERROR] Counts row {i} has fewer than 2 columns.", file=sys.stderr)
                sys.exit(1)
            tx = row[0]
            counts = row[1:]
            rows.append((tx, counts))

    if not rows:
        print("[ERROR] Counts file has no data rows.", file=sys.stderr)
        sys.exit(1)

    return sample_headers, rows

def ensure_sets_match(tx2gene, counts_rows):
    set_A = set(tx2gene.keys())
    set_B = set(tx for tx, _ in counts_rows)

    if set_A != set_B:
        only_in_A = sorted(set_A - set_B)
        only_in_B = sorted(set_B - set_A)
        msg_lines = ["[ERROR] Transcript ID sets do not match between GTF(attributes) and counts."]
        if only_in_A:
            msg_lines.append(f"  - Present only in GTF (A): {len(only_in_A)} examples: {only_in_A[:10]}")
        if only_in_B:
            msg_lines.append(f"  - Present only in counts (B): {len(only_in_B)} examples: {only_in_B[:10]}")
        msg_lines.append("This indicates some transcripts appear in only A or only B. "
                         "Please resolve the mismatch in your inputs.")
        print("\n".join(msg_lines), file=sys.stderr)
        sys.exit(1)

def write_output(output_path, sample_headers, counts_rows, tx2gene):
    """
    DRIMSeq 形式のTSV:
      gene_id  feature_id  sample1  sample2  ...
      (fileBの行順を維持)
    """
    with open(output_path, "w", encoding="utf-8", newline="") as f:
        writer = csv.writer(f, delimiter="\t")
        header = ["gene_id", "feature_id"] + sample_headers
        writer.writerow(header)

        for tx, counts in counts_rows:
            gene = tx2gene.get(tx)
            if gene is None:
                # 通常ここには来ない（事前一致チェック済み）
                print(f"[WARN] tx {tx!r} not found in mapping; skipping.", file=sys.stderr)
                continue
            writer.writerow([gene, tx] + counts)

def main():
    args = parse_args()

    tx2gene = build_tx_to_gene_map(args.gtf_tsv, args.gene_key, args.tx_key)
    sample_headers, counts_rows = read_counts(args.counts)
    ensure_sets_match(tx2gene, counts_rows)
    write_output(args.output, sample_headers, counts_rows, tx2gene)

    print(f"[OK] Wrote DRIMSeq counts table: {args.output}", file=sys.stderr)

if __name__ == "__main__":
    main()




import subprocess
import sys
# arg1 - r script for execution of edgeR
# arg2 - meta data
# arg3 - count data
# arg4 - output (tests)
# arg5 - output (CPM)

def de(r_script, meta_data, count_data, result_test, result_cpm):
    # Rスクリプトを実行
    subprocess.run(["Rscript", r_script, meta_data, count_data, result_test, result_cpm])

if __name__ ==  "__main__":
    de(sys.argv[1], sys.argv[2], sys.argv[3], sys.argv[4], sys.argv[5])
    print("edgeR analysis done!")



""" 目的
    1. TargetP の結果を回収
    2. GTFprepフォーマットで出力
"""

import re
import pandas as pd
import argparse
import sys
import csv


### TargetP2 (mTPを含む)
#       # TargetP-2.0   Organism: Non-Plant     Timestamp: 20250314183052
#       # ID    		Prediction      noTP    	SP      	mTP     	CS Position
#       ENST00000465205.2       mTP     	0.296826        0.000283        0.702891        CS pos: 25-26. VMG-KQ. Pr: 0.1151
#       ENST00000261405.10      SP      	0.000257        0.999722        0.000021        CS pos: 22-23. TLC-AE. Pr: 0.8440


def get_cleavage_site(text):
    # 正規表現でCS posの2つの数字を取り出す
    match = re.search(r'CS pos:\s*(\d+)-(\d+)', text)
    if match:
        cs_start = int(match.group(1))
        cs_end = int(match.group(2))
    return cs_start


def get_csprob(text):
    # 正規表現でCS posの確率を取り出す
    match = re.search(r'Pr:\s*([0-9]*\.[0-9]+)', text)
    if match:
        cs_prob = float(match.group(1))
    return cs_prob


def parse_result(file):
    df = []
    with open(file, "r") as f:
        for line in f:
            line = line.strip()
            if line.startswith("#"):
                continue

            if line.startswith("<query>"):
                continue

            if not line:
                continue

            cols = line.split('\t')
            if len(cols) != 6:
                continue

            if cols[1] != "mTP":
            # mTP prediction のみ処理
                continue

            if "?" in cols[5]:
                continue
            # Cleavage site
            cleavage_site = get_cleavage_site(cols[5])
            cleavage_prob = get_csprob(cols[5])

            df.append([
                cols[0],		# 転写産物名
                1,			# from
                cleavage_site,		# to
                float(cols[4]),		# mTP prob.
                cleavage_prob,		# cleavage site prob.
            ])
    return pd.DataFrame(df, columns=["Transcript", "from", "to", "mTP_prob", "cleavage_prob"])


def filter_threshold(df, CUTOFF, SINGLE):
    # CUTOFF 未満の結果をフィルター（デフォルト：0.9）
    df_filtered = df[df["mTP_prob"] >= CUTOFF]

    # 転写産物につき1つの結果だけを残す（デフォルト：しない）
    df_filtered = df_filtered.loc[df_filtered.groupby('Transcript')['cleavage_prob'].idxmax()]
    return df_filtered


def save_gtfprep(filtered_df, output_path, feature, source):
    """
    Parameters:
        filtered_df (pd.DataFrame): フィルタ済みデータフレーム
        output_path (str): 出力ファイルパス（TSV）
        feature (str): feature 列に使う固定文字列
        source (str): source 列に使う固定文字列
    """
    # attributes列を作成
    def format_attributes(row):
        return (
            'mTP ' + '"'+str(row["mTP_prob"]).replace('"', '')+'"' + '; '
            'CS_prob ' + '"'+str(row["cleavage_prob"]).replace('"', '')+'"' + '; '
        )

    attributes = filtered_df.apply(format_attributes, axis=1)

    output_df = pd.DataFrame({
        "#Transcript": filtered_df["Transcript"],
        "tx_from": filtered_df["from"],
        "tx_to": filtered_df["to"],
        "amino_acid": "amino_acid",
        "feature": feature,
        "source": source,
        "attributes": attributes
    })

    # 保存
    output_df.to_csv(output_path, sep="\t", index=False, quoting=csv.QUOTE_NONE)
    print(">>> Output file: ", output_path)


def str2bool(value):
    if isinstance(value, bool):
        return value
    if value.lower() == "true":
        return True
    elif value.lower() == "false":
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected: 'true' or 'false'")


def main():
    parser = argparse.ArgumentParser(description="Make gtfprep format from TargetP results")
    parser.add_argument("output", help="Path to TargetP output file")
    parser.add_argument("feature", help="feature name column")
    parser.add_argument("source", help="source name column")
    parser.add_argument("--cutoff", type=float, default=0.9, help="Cutoff of mTP prob (default: 0.9)")
    parser.add_argument("--single", type=str2bool, default=False, help="true or false. True if you keep only one detected result for each transcript. (default: False)")

    args = parser.parse_args()

    # 引数
    nls_output = args.output
    feature = args.feature
    source = args.source
    cutoff = args.cutoff
    single = args.single

    # メイン処理
    output_path = nls_output + ".gtfprep"
    df = parse_result(nls_output)
    filtered_df = filter_threshold(df, cutoff, single)
    save_gtfprep(filtered_df, output_path, feature, source)

if __name__ == "__main__":
    main()




# handler function
# src/isoespy/cli.py
import sys

from .isoespy_de import main as de_main
from .isoespy_diu import main as diu_main
from .isoespy_ff import main as ff_main
from .isoespy_edger import main as edger_main
from .run_DRIMSeq_stageR import main as drimseq_main
from .isoespy_makefa import main as makefa_main
from .isoespy_makegtf import main as makegtf_main
from .isoespy_orfpred import main as orfpred_main


COMMANDS = {
    "de": de_main,
    "diu": diu_main,
    "ff": ff_main,
    "edger": edger_main,
    "drim": drimseq_main,
    "fas": makefa_main,
    "gtf": makegtf_main,
    "orf": orfpred_main,
}


def _print_help() -> None:
    print("Usage: isoespy <command> [options]\n")
    print("Available commands:")
    print("$ isoespy  de [options]       Differential transcript expression visualization")
    print("isoespy de visualizes the exon structures of each isoform of the target gene along with the expression distributions of the two groups and the results of the differential expression (DE) analysis.")
    print("")

    print("$ isoespy  diu [options]      Differential isoform usage visualization")
    print("isoespy diu visualizes the exon structures of each isoform of the target gene, the isoform usage proportions within each group, and the changes in usage between the two groups.")
    print("")

    print("$ isoespy  ff [options]       Functional feature visualization")
    print("isoespy ff visualizes the exon structures and functional features of each isoform of the target gene.")
    print("")

    print("$ isoespy  edger [options]    edgeR-based transcript-level DE analysis")
    print("isoespy edger is a wrapper for edgeR. This command performs transcript-level differential expression analysis using the edgeR R package.")
    print("")

    print("$ isoespy  drim [options]     DRIMSeq and stageR-based DIU analysis")
    print("isoespy drim is a wrapper for DRIMSeq and stageR. It performs gene-level differential isoform usage (DIU) analysis using these R packages.")
    print("")

    print("$ isoespy  fas [options]      Handle FASTA file")
    print("isoespy fas extracts FASTA sequences from a GTF File. To extract nucleotide sequences for transcript models from an input GTF file, specify feature exon and type nucleotide. If you wish to obtain amino acid sequences instead, specify feature CDS and type amino_acid. Because sequence retrieval is performed using the pysam library, the reference genome index file (.fai) must be located in the same directory as the reference FASTA file.")
    print("")

    print("$ isoespy  gtf [options]      Handle GTF file")
    print("isoespy gtf integrates GTFprep annotations into a GTF file. By providing a GTFprep file that contains functional annotation information, isoespy gtf generates a new GTF file in which these annotations are appended to the corresponding transcripts.")
    print("")

    print("$ isoespy  orf [options]      ORF prediction")
    print("isoespy orf predicts ORFs for transcripts that lack CDS annotations in GTF file. For any transcript in the input GTF that does not contain a CDS feature, isoespy uses CPAT and TransDecoder to predict coding sequences and appends the corresponding CDS annotations to the GTF. The hexamer and model options in the following command correspond to human. When analyzing other organisms, please provide the appropriate pretrained datasets for the target species.")
    print("")

def main() -> int:
    """Entry point for the `isoespy` command with subcommands."""

    # サブコマンドが指定されていない / help のとき
    if len(sys.argv) < 2 or sys.argv[1] in ("-h", "--help"):
        _print_help()
        return 0

    cmd = sys.argv[1]

    # サブコマンド名を argv から取り除き、
    # 残りをそのまま各 main() に渡せるようにする
    sys.argv = [sys.argv[0]] + sys.argv[2:]

    if cmd in COMMANDS:
        # 各 isoespy_*.py の main() は従来通り sys.argv から引数を読むはずなので、
        # ここでは単純に呼び出すだけでよい
        return COMMANDS[cmd]()
    else:
        print(f"Unknown command: {cmd}\n")
        _print_help()
        return 1


if __name__ == "__main__":
    raise SystemExit(main())





""" 目的
    1. NESmapper の結果を回収
    2. GTFprepフォーマットで出力
"""

import pandas as pd
import argparse
import sys
import csv

### NESmapper
#	<query>			<pos>	<nes>			<score>		<nes-class>
#	AADAT_novel_tx_1	216	IYELARKYDFLIIE		18		class_1a
#	ABCC2_novel_tx_1	566	RPELDLVLRGITCD		12.6		class_1a
#	ABCE1_novel_tx_1	27	KSCPVVRMGKLCIE		12.6		class_1b


def parse_result(file):
    df = []
    with open(file, "r") as f:
        for line in f:
            line = line.strip()
            if line.startswith("#"):
                continue

            if line.startswith("<query>"):
                continue

            if not line:
                continue

            cols = line.split('\t')
            if len(cols) != 5:
                continue

            df.append([
                cols[0],			# 転写産物名
                int(cols[1]),			# from
                int(cols[1])+len(cols[2])-1,	# to
                float(cols[3]),			# score
                cols[4],			# nes-class
            ])
    return pd.DataFrame(df, columns=["Transcript", "from", "to", "score", "nes-class"])


def filter_threshold(df, CUTOFF, SINGLE):
    # CUTOFF 未満の結果をフィルター（デフォルト：12）
    df_filtered = df[df["score"] >= CUTOFF]

    # 転写産物につき1つの結果だけを残す（デフォルト：しない）
    df_filtered = df_filtered.loc[df_filtered.groupby('Transcript')['score'].idxmax()]
    return df_filtered


def save_gtfprep(filtered_df, output_path, feature, source):
    """
    Parameters:
        filtered_df (pd.DataFrame): フィルタ済みデータフレーム
        output_path (str): 出力ファイルパス（TSV）
        feature (str): feature 列に使う固定文字列
        source (str): source 列に使う固定文字列
    """
    # attributes列を作成
    def format_attributes(row):
        return (
            'score ' + '"'+str(row["score"]).replace('"', '')+'"' + '; '
            'new-class ' + '"'+str(row["nes-class"]).replace('"', '')+'"' + '; '
        )

    attributes = filtered_df.apply(format_attributes, axis=1)

    output_df = pd.DataFrame({
        "#Transcript": filtered_df["Transcript"],
        "tx_from": filtered_df["from"],
        "tx_to": filtered_df["to"],
        "amino_acid": "amino_acid",
        "feature": feature,
        "source": source,
        "attributes": attributes
    })

    # 保存
    output_df.to_csv(output_path, sep="\t", index=False, quoting=csv.QUOTE_NONE)
    print(">>> Output file: ", output_path)


def str2bool(value):
    if isinstance(value, bool):
        return value
    if value.lower() == "true":
        return True
    elif value.lower() == "false":
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected: 'true' or 'false'")


def main():
    parser = argparse.ArgumentParser(description="Make gtfprep format from NESmapper results")
    parser.add_argument("output", help="Path to NESmapper output file")
    parser.add_argument("feature", help="feature name column")
    parser.add_argument("source", help="source name column")
    parser.add_argument("--cutoff", type=float, default=12, help="Cutoff of score (default: 12)")
    parser.add_argument("--single", type=str2bool, default=False, help="true or false. True if you keep only one detected result for each transcript. (default: False)")

    args = parser.parse_args()

    # 引数
    nls_output = args.output
    feature = args.feature
    source = args.source
    cutoff = args.cutoff
    single = args.single

    # メイン処理
    output_path = nls_output + ".gtfprep"
    df = parse_result(nls_output)
    filtered_df = filter_threshold(df, cutoff, single)
    save_gtfprep(filtered_df, output_path, feature, source)

if __name__ == "__main__":
    main()




import shutil
import subprocess
import sys
import argparse
import os
import csv
import re
import pandas as pd
from collections import OrderedDict

def run_r_script(
    r_script_path,
    input1,
    input2,
    input3,
    output1,
    output2
):
    """
    Run an R script via subprocess with 3 input files and 2 output files.

    Parameters
    ----------
    r_script_path : str
        Path to the R script (e.g., "script_A.R")
    input1, input2, input3 : str
        Paths to input files
    output1, output2 : str
        Paths to output files
    """

    # check that Rscript is available
    if not shutil.which("Rscript"):
        raise EnvironmentError("Rscript not found in PATH. Please ensure R is installed and accessible.")

    # construct command
    cmd = [
        "Rscript",
        r_script_path,
        input1,
        input2,
        input3,
        output1,
        output2
    ]

    print("Running command:")
    print(" ".join(cmd))

    try:
        subprocess.run(cmd, check=True)
        print("R script executed successfully.")
    except subprocess.CalledProcessError as e:
        print(f"Error: R script failed with exit code {e.returncode}")
        sys.exit(1)


def parse_attributes(attr_field):
    """
      - key "value"; key "value";
      - key=value;key=value;
      - key value; のような形式
    """
    attr = {}
    if attr_field is None:
        return attr
    s = attr_field.strip().strip(";")
    if not s:
        return attr

    # ; で分割（連続セミコロンや末尾空要素を無視）
    parts = re.split(r';\s*', s)
    for part in parts:
        part = part.strip()
        if not part:
            continue
        key = None
        val = None

        if '=' in part:
            key, val = part.split('=', 1)
        else:
            # key "value" あるいは key value を拾う
            m = re.match(r'(\S+)\s+"?(.*?)"?$', part)
            if m:
                key, val = m.group(1), m.group(2)

        if key is not None:
            key = key.strip()
            if val is not None:
                val = val.strip().strip('"')
            attr[key] = val
    return attr

def build_tx_to_gene_map(gtf_tsv_path, gene_key, tx_key):
    """
    第9列 attributes から gene_key と tx_key を取り出し、
    転写産物ID -> 遺伝子ID の辞書を作る。
    同一転写産物IDが複数行に出ても最初の対応を採用（2回目以降は無視）。
    """
    tx2gene = OrderedDict()
    line_no = 0
    with open(gtf_tsv_path, "r", encoding="utf-8-sig", newline="") as f:
        reader = csv.reader(f, delimiter="\t")
        for row in reader:
            line_no += 1
            if not row or len(row) < 9:
                # 不完全行はスキップ
                continue
            attributes_field = row[8]
            attrs = parse_attributes(attributes_field)

            tx = attrs.get(tx_key)
            gene = attrs.get(gene_key)

            if tx is None or gene is None:
                # 必要キーが無い行はスキップ（構造上、exon行など全てに両方あるとは限らないため）
                continue

            if tx not in tx2gene:
                tx2gene[tx] = gene
            else:
                # 2回目以降は飛ばす（ただし gene が不一致なら警告）
                if tx2gene[tx] != gene:
                    print(
                        f"[WARN] Transcript {tx!r} already mapped to {tx2gene[tx]!r}, "
                        f"but found another gene {gene!r} at line {line_no}. Keeping the first mapping.",
                        file=sys.stderr
                    )
    if not tx2gene:
        print("[ERROR] No (transcript_id, gene_id) pairs were extracted from attributes. "
              f"Check that 9th column contains attributes and keys '{tx_key}', '{gene_key}'.",
              file=sys.stderr)
        sys.exit(1)
    return tx2gene

def read_counts(counts_path):
    """
    counts TSV を読み込み：
      - 1行目: ヘッダー（col1=tx、col2+=サンプル名）
      - 2行目以降: tx, sample1_count, sample2_count, ...
    戻り値:
      header_samples: [sample1, sample2, ...]
      rows: [(tx, [c1, c2, ...]), ...]  ※入力の順序維持
    """
    rows = []
    with open(counts_path, "r", encoding="utf-8-sig", newline="") as f:
        reader = csv.reader(f, delimiter="\t")
        try:
            header = next(reader)
        except StopIteration:
            print("[ERROR] Counts file is empty.", file=sys.stderr)
            sys.exit(1)
        if len(header) < 2:
            print("[ERROR] Counts header must have at least two columns: tx and >=1 sample.", file=sys.stderr)
            sys.exit(1)

        # 1列目は tx、2列目以降はサンプル名
        sample_headers = header[1:]

        for i, row in enumerate(reader, start=2):
            if not row:
                continue
            if len(row) < 2:
                print(f"[ERROR] Counts row {i} has fewer than 2 columns.", file=sys.stderr)
                sys.exit(1)
            tx = row[0]
            counts = row[1:]
            rows.append((tx, counts))

    if not rows:
        print("[ERROR] Counts file has no data rows.", file=sys.stderr)
        sys.exit(1)

    return sample_headers, rows


def ensure_sets_match(tx2gene, counts_rows):
    """
    GTF は発現量データをグループ化するために使うだけなので、
    GTF にしかない transcript は許容する。
    ただし、counts 側にある transcript が GTF に無いのはエラーにする。
    """
    gtf_txs    = set(tx2gene.keys())               # GTFに書いてあるtranscript
    counts_txs = set(tx for tx, _ in counts_rows)  # countsに登場するtranscript

    # counts にあるのに GTF に無いものだけチェック
    missing_in_gtf = sorted(counts_txs - gtf_txs)

    if missing_in_gtf:
        msg_lines = ["[ERROR] Some transcripts in counts were not found in GTF (these cannot be grouped by gene)."]
        msg_lines.append(f"  - Missing in GTF: {len(missing_in_gtf)} examples: {missing_in_gtf[:10]}")
        msg_lines.append("Please make sure every transcript in the expression data exists in the GTF.")
        print("\n".join(msg_lines), file=sys.stderr)
        sys.exit(1)

    # 逆に「GTFにだけあるもの」は何もしないでOK


def write_output(output_path, sample_headers, counts_rows, tx2gene):
    """
    DRIMSeq 形式のTSV:
      gene_id  feature_id  sample1  sample2  ...
      (fileBの行順を維持)
    """
    with open(output_path, "w", encoding="utf-8", newline="") as f:
        writer = csv.writer(f, delimiter="\t")
        header = ["gene_id", "feature_id"] + sample_headers
        writer.writerow(header)

        for tx, counts in counts_rows:
            gene = tx2gene.get(tx)
            if gene is None:
                # 通常ここには来ない（事前一致チェック済み）
                print(f"[WARN] tx {tx!r} not found in mapping; skipping.", file=sys.stderr)
                continue
            writer.writerow([gene, tx] + counts)


def merge_results(pval_file, prop_file, sample_file, output_file):
    NORM = "dot2dash"
    # --- load ---
    df_stage = pd.read_csv(pval_file, sep="\t")
    df_prop  = pd.read_csv(prop_file, sep="\t")
    df_smp   = pd.read_csv(sample_file, sep="\t")

    # --- rename txID->feature_id if needed ---
    if "txID" in df_stage.columns:
        df_stage = df_stage.rename(columns={"txID": "feature_id"})

    # --- choose id columns and sample columns ---
    id_cols = [c for c in ["gene_id", "feature_id"] if c in df_prop.columns]
    if not set(["gene_id","feature_id"]).issubset(df_prop.columns):
        sys.exit("ERROR: proportions TSV must contain gene_id and feature_id columns.")

    sample_cols = [c for c in df_prop.columns if c not in id_cols]
    if len(sample_cols) == 0:
        sys.exit("ERROR: No sample columns detected in proportions TSV.")

    # --- normalize sample IDs consistently with samples.tsv ---
    def normalize(s: str) -> str:
        if NORM == "dot2dash":
            return s.replace(".", "-")
        elif NORM == "dash2dot":
            return s.replace("-", ".")
        return s

    norm_map = {col: normalize(col) for col in sample_cols}

    # melt to long
    df_long = df_prop.melt(
        id_vars=id_cols,
        value_vars=sample_cols,
        var_name="sample_col",
        value_name="proportion"
    )
    df_long["sample_id"] = df_long["sample_col"].map(norm_map)

    # --- join group info ---
    if not set(["sample_id","group"]).issubset(df_smp.columns):
        sys.exit("ERROR: samples TSV must contain 'sample_id' and 'group' columns.")
    df_long = df_long.merge(df_smp[["sample_id","group"]], on="sample_id", how="left")

    # quick sanity check
    n_unmatched = df_long["group"].isna().sum()
    if n_unmatched > 0:
        uniq_unmatched = sorted(df_long.loc[df_long["group"].isna(), "sample_id"].unique().tolist())
        print(f"[WARN] {n_unmatched} rows had no group (first few unmatched IDs: {uniq_unmatched[:10]})", file=sys.stderr)

    # --- compute group-wise mean proportions ---
    grp_cols = id_cols + ["group"]
    mean_prop = (df_long.dropna(subset=["group"])
                 .groupby(grp_cols, as_index=False)["proportion"]
                 .mean())

    # pivot to wide: prop_<group>
    wide = (mean_prop
            .pivot_table(index=id_cols, columns="group", values="proportion")
            .reset_index())

    # --- merge to stageR (by feature_id) ---
    if "feature_id" not in df_stage.columns:
        sys.exit("ERROR: stageR TSV must contain column 'feature_id' (or txID).")
    
    if "gene_id" in wide.columns:
        wide_for_merge = wide.drop(columns=["gene_id"])
    else:
        wide_for_merge = wide

    out = df_stage.merge(wide_for_merge, on="feature_id", how="left")

    current_cols = list(out.columns)
    
    # 基本の4列（gene_id, tx_id, gene_p, tx_p）をリネーム
    if len(current_cols) >= 4:
        current_cols[0] = "gene_id"
        current_cols[1] = "tx_id"
        current_cols[2] = "gene_p"
        current_cols[3] = "tx_p"
    
    # 5列目以降（prop列）を prop_0, prop_1 ... にリネーム
    for i in range(4, len(current_cols)):
        current_cols[i] = f"prop_{i-4}"
    
    out.columns = current_cols

    # --- write ---
    out.to_csv(output_file, sep="\t", index=False)
    print(f"Done: {output_file}")


def main():
    # Arguments
    parser = argparse.ArgumentParser(description="Run an R script with 3 input and 2 output files.")
    parser.add_argument("--r_script", required=True, help="Path to the R script (e.g., script_A.R)")
    parser.add_argument("--count", required=True)
    parser.add_argument("--sample", required=True)
    parser.add_argument("--config", required=True)
    parser.add_argument("--output_pval", required=True)
    parser.add_argument("--output_prop", required=True)
    parser.add_argument("--gtf", required=True)
    parser.add_argument("--gene_key", default="gene_symbol", help="Attribute key for gene id (default: gene_symbol)")
    parser.add_argument("--tx_key", default="transcript_id", help="Attribute key for transcript id (default: transcript_id)")
    parser.add_argument("--OUT", default="result.tsv", help="Output TSV")

    args = parser.parse_args()

    # Prepare DRIMSeq input files
    tx2gene = build_tx_to_gene_map(args.gtf, args.gene_key, args.tx_key)
    sample_headers, counts_rows = read_counts(args.count)
    ensure_sets_match(tx2gene, counts_rows)
    drimseq_count = os.path.dirname(args.sample)+"/"+os.path.basename(args.count)+".DRIMSeq.tsv"
    write_output(drimseq_count, sample_headers, counts_rows, tx2gene)
    print(f"[OK] Wrote DRIMSeq counts table: {drimseq_count}", file=sys.stderr)

    # DRIMSeq + stageR
    run_r_script(
        args.r_script,
        drimseq_count,
        args.sample,
        args.config,
        args.output_pval,
        args.output_prop
    )

    # Merge results
    merge_results(args.output_pval, args.output_prop, args.sample, args.OUT)


if __name__ == "__main__":
    main()





import argparse
import subprocess
try:
    from run_edgeR_n_get_results import de
except ModuleNotFoundError:
    from isoespy.run_edgeR_n_get_results import de

def main(args=None):
    # parser
    parser = argparse.ArgumentParser(description='isoespy_edger')
    parser.add_argument('-r', '--r_script', type=str, default="edgeR_script.R", help='R script for edgeR')
    parser.add_argument('-m', '--meta_data', required=True, type=str, default=None, help='metadata')
    parser.add_argument('-c', '--count_data', required=True, type=str, default=None, help='count expression data')
    parser.add_argument('-o1', '--output_de', required=True, type=str, default=None, help='result of edgeR DE analysis')
    parser.add_argument('-o2', '--output_cpm', required=True, type=str, default=None, help='result of edgeR CPM expression data')

    args = parser.parse_args()
    r_script = args.r_script
    meta_data = args.meta_data
    count_data = args.count_data
    det_data = args.output_de
    cpm_data = args.output_cpm

    # DE analysis
    de(r_script, meta_data, count_data, det_data, cpm_data)

    # adjustment
    det_data_reorder = det_data + ".reorder"
    subprocess.run([
        "awk", "-F,",
        '{print $6","$1","$5","$2","$3","$4}',
        det_data
        ], stdout=open(det_data_reorder, "w"))
    subprocess.run(["awk", "NR==1{$0=\"#\"$0}1", det_data_reorder], stdout=open("temp.txt", "w"))
    subprocess.run(["mv", "temp.txt", det_data_reorder])


if __name__ == "__main__":
    main()





def calc_populations(isoform_model):
    """ P_ranges と P_components を計算する """
    # 辞書から範囲を展開し、ソート
    indexed_ranges = [(start, end, key, i) for key, ranges in isoform_model.items() for i, (start, end) in enumerate(ranges)]
    indexed_ranges.sort()

    merged = []
    merged_indices = []
    
    for start, end, source, index in indexed_ranges:
        if (not merged) or (merged[-1][1] < start-1):
            # 新規P (現時点の最右populationとオーバーラップしないため新規)
            merged.append((start, end))
            merged_indices.append({source: [index]})
        else:
            # 既存P に追加
            merged[-1] = (merged[-1][0], max(merged[-1][1], end))
            if source in merged_indices[-1]:
                merged_indices[-1][source].append(index)
            else:
                merged_indices[-1][source] = [index]
    
    P_ranges = {i+1: merged[i] for i in range(len(merged))}
    P_components = {i+1: merged_indices[i] for i in range(len(merged))}
    for tx in isoform_model:
        for i in P_components:
            if not tx in P_components[i]:
                P_components[i][tx] = []
 
    return P_ranges, P_components


def calc_introns(p_ranges):
    sorted_ranges = sorted(p_ranges.values())
    gaps = {}
    
    gap_start = None
    gap_index = 1
    
    for i in range(len(sorted_ranges) - 1):
        gap_start = sorted_ranges[i][1] + 1
        gap_end = sorted_ranges[i+1][0] - 1
        
        if gap_start <= gap_end:
            gaps[gap_index] = (gap_start, gap_end)
            gap_index += 1
    
    return gaps


def calc_migration_distance(i_ranges, rate):
    if rate < 1:
        # 方法1: イントロン幅を 100*rate % に圧縮
        p_left = {}
        p_left[1] = 0
        for i in i_ranges:
            p_left[i+1] = 0
            # 左に詰める分だけ加算
            for j in range(1, i+1):
                p_left[i+1] += (i_ranges[j][1]-i_ranges[j][0]+1)*(1-rate)
    else:
        # 方法2: イントロン幅を rate [nt] に固定
        p_left = {}
        p_left[1] = 0
        for i in i_ranges:
            p_left[i+1] = 0
            # 左に詰める分だけ加算
            for j in range(1, i+1):
                width = i_ranges[j][1]-i_ranges[j][0]+1
                p_left[i+1] += width - rate
    return p_left


def migrate_to_left(isoform_model, p_left, p_components):
    """ isoform_model: 元の座標
        p_left:        左づめの分量
        p_components:  対象となるエクソン
    """
    moved_model = {tx: [] for tx in isoform_model}
    for tx in isoform_model:
        number_of_exons = len(isoform_model[tx])
        for i in range(number_of_exons):
            # txの第iエクソン (0, 1, 2,...)
            for j in p_components:
                # Pjに属しているか
                if i in p_components[j][tx]:
                    moved_model[tx].append((isoform_model[tx][i][0]-p_left[j], isoform_model[tx][i][1]-p_left[j]))
                    break
    return moved_model


def intronCompression(isoform_model, rate=100):
    P_ranges, P_components = calc_populations(isoform_model)
    I_ranges = calc_introns(P_ranges)
    P_left = calc_migration_distance(I_ranges, rate)
    moved_isoform_model = migrate_to_left(isoform_model, P_left, P_components)
    #MIN = float('inf')
    #MAX = -float('inf')
    #for tx in moved_isoform_model:
    #    if moved_isoform_model[tx][0][0] < MIN:
    #        MIN = moved_isoform_model[tx][0][0]
    #    if MAX < moved_isoform_model[tx][-1][1]:
    #        MAX = moved_isoform_model[tx][-1][1]
    return moved_isoform_model


if __name__ == "__main__":
    # テストケース
    isoform_model = {
        "tx1": [(30,60), (80,110), (150,180), (200,210), (250,270)],
        "tx2": [(50,90), (150,180), (230,230), (260,280)],
        "tx3": [(70,90), (150,170), (250,280)]
    }

    P_ranges, P_components = calc_populations(isoform_model)
    I_ranges = calc_introns(P_ranges)
    P_left = calc_migration_distance(I_ranges, 0.1)
    print("P_ranges", P_ranges)
    print("P_components", P_components)
    print("I_ranges", I_ranges)
    print("P_left", P_left)
    print("isoform_model", isoform_model)
    moved_isoform_model = migrate_to_left(isoform_model, P_left, P_components)
    print("moved_isoform_model", moved_isoform_model)




""" 目的
    1. hmmscan の結果から有意なドメインを回収
    2. transcript-based annotation のフォーマット（GTF化の前段階）で出力
"""

import pandas as pd
import argparse
import sys
import csv

# #                                                                            --- full sequence --- -------------- this domain -------------   hmm coord   ali coord   env coord
# # target name        accession   tlen query name           accession   qlen   E-value  score  bias   #  of  c-Evalue  i-Evalue  score  bias  from    to  from    to  from    to  acc description of target
# #------------------- ---------- ----- -------------------- ---------- ----- --------- ------ ----- --- --- --------- --------- ------ ----- ----- ----- ----- ----- ----- ----- ---- ---------------------
# 7tm_2                PF00002.29   250 ENST00000283303.3    -            695   2.9e-33  115.5  23.6   1   1   6.9e-37   5.5e-33  114.7  23.6     3   250   400   648   398   648 0.86 7 transmembrane receptor (Secretin family)

# |
# |
# |
# V

# #transcript_id	from	to	nt/aa		feature_name		source		attribute_field
# ADGRF4_novel_tx_1	252     500	amino_acid	protein_domain (引数)	Pfam (引数)	domain_id "PF22691.2"; domain_name "Thiolase_C_1"; evalue "4.3e-36";


# hmmscan の domtbl をパース
def parse_hmmscan_result(file, coord):
    if coord == "ali":	# ali
        C = 17
    else:		# env
        C = 19

    df = []
    with open(file, "r") as f:
        for line in f:
            if line.startswith("#"):
                continue
            cols = line.split()
            df.append([
                cols[3],			# 転写産物名
                cols[0],			# PFAM shortname
                cols[1],			# PFAM accession
                float(cols[12]),		# i-Evalue
                int(cols[2]),			# target length
                int(cols[15]), int(cols[16]),	# hmm from, hmm to
                int(cols[C]), int(cols[C+1]),	# tx from, tx to
                float(cols[21]),		# The mean posterior probability of aligned residues in the MEA alignment
                int(cols[5]),			# CDS total length
            ])
    return pd.DataFrame(df, columns=["Transcript", "Shortname", "Pfam_accession", "Evalue", "Domain_len", "hmm_from", "hmm_to", "tx_from", "tx_to", "Accuracy", "CDS_total_len"])


def remove_overlapping_domains(df):
    result_list = []
    for transcript, group in df.groupby("Transcript"):
        group = group.sort_values("Evalue").reset_index(drop=True)
        kept = []

        for i, row in group.iterrows():
            # 現在の領域
            from_i, to_i = row["tx_from"], row["tx_to"]
            overlap = False

            for kept_row in kept:
                from_k, to_k = kept_row["tx_from"], kept_row["tx_to"]
                # 重なり判定（両端含む）
                if not (to_i < from_k or from_i > to_k):
                    overlap = True
                    break

            if not overlap:
                kept.append(row)

        result_list.extend(kept)

    return pd.DataFrame(result_list)


def save_domain_annotation_tsv(filtered_df, output_path, feature, source):
    """
    指定された filtered_df からドメイン注釈ファイルを作成して保存する。

    Parameters:
        filtered_df (pd.DataFrame): ドメイン予測結果のフィルタ済みデータフレーム
        output_path (str): 出力ファイルパス（TSV）
        feature (str): feature 列に使う固定文字列
        source (str): source 列に使う固定文字列
    """
    # attributes列を作成
    def format_attributes(row):
        return (
            'domain_id ' + '"'+str(row["Pfam_accession"]).replace('"', '')+'"' + '; '
            'domain_name ' + '"'+str(row["Shortname"]).replace('"', '')+'"' + '; '
            'evalue ' + '"'+str(row["Evalue"]).replace('"', '')+'"' + ';'
        )

    attributes = filtered_df.apply(format_attributes, axis=1)

    output_df = pd.DataFrame({
        "#Transcript": filtered_df["Transcript"],
        "tx_from": filtered_df["tx_from"],
        "tx_to": filtered_df["tx_to"],
        "amino_acid": "amino_acid",
        "feature": feature,
        "source": source,
        "attributes": attributes
    })

    # 保存
    output_df.to_csv(output_path, sep="\t", index=False, quoting=csv.QUOTE_NONE)


def filter_hmmscan_result(df, evalue_threshold, accuracy_threshold, coverage_threshold, output_path, feature, source, overlap_remover):
    """
    HMMER hmmscanの出力データをフィルタリングし、条件を満たすドメインのみを保存する。

    Parameters:
        evalue_threshold (float): E-valueのカットオフ（デフォルト: 1e-5）
        accuracy_threshold (float): Accuracyのカットオフ（デフォルト: 0.7）
        coverage_threshold (float): アライメントカバレッジのカットオフ（デフォルト: 0.5）
    """

    # アライメント割合を計算
    df["alignment_ratio"] = (df["hmm_to"] - df["hmm_from"] + 1) / df["Domain_len"]

    # (1) Evalue < evalue_threshold
    filtered_df = df[df["Evalue"] < evalue_threshold].copy()

    # (2) アライメント割合 > alignment_ratio_threshold
    filtered_df = filtered_df[filtered_df["alignment_ratio"] > coverage_threshold]

    # (3) Accuracy > accuracy_threshold
    filtered_df = filtered_df[filtered_df["Accuracy"] > accuracy_threshold]

    # (4) (overlap_remover == True の場合のみ) Transcriptごとに予測領域の重なるドメインはEvalueが最小のものに絞る
    if overlap_remover:
        filtered_df = remove_overlapping_domains(filtered_df)

    # 結果を保存
    output_path = output_path+".gtfprep"
    save_domain_annotation_tsv(filtered_df, output_path, feature, source)


def str2bool(value):
    if isinstance(value, bool):
        return value
    if value.lower() == "true":
        return True
    elif value.lower() == "false":
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected: 'true' or 'false'")


def main():
    parser = argparse.ArgumentParser(description="Filter hmmscan results")
    parser.add_argument("domtbl", help="Path to domtbl file")
    parser.add_argument("feature", help="feature name column")
    parser.add_argument("source", help="source name column")
    parser.add_argument("--evalue", type=float, default=1e-5, help="E-value threshold (default: 1e-5)")
    parser.add_argument("--coverage", type=float, default=0.5, help="Minimum alignment coverage ratio of the query to the full length of  domain model (default: 0.5)")
    parser.add_argument("--acc", type=float, default=0.7, help="Accuracy threshold (default: 0.7)")
    parser.add_argument("--filt_overlap", type=str2bool, default=False, help="true or false. true if overlapped domains are not allowed. (default: False)")

    args = parser.parse_args()

    # 必須引数
    domtbl = args.domtbl
    feature = args.feature
    source = args.source

    # オプション
    evalue_threshold = args.evalue
    coverage_threshold = args.coverage
    acc_threshold = args.acc
    coord = "ali"
    overlap_remover = args.filt_overlap

    # メイン処理
    df = parse_hmmscan_result(domtbl, coord)
    filter_hmmscan_result(df, evalue_threshold, acc_threshold, coverage_threshold, domtbl, feature, source, overlap_remover)

if __name__ == "__main__":
    main()

